# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import math
import numpy as np
import quaternion
import scri
import spherical_functions as sf
import warnings


def modes_constructor(constructor_statement, data_functor, **kwargs):
    """WaveformModes object filled with data from the input functor

    Additional keyword arguments are mostly passed to the WaveformModes initializer, though some more reasonable
    defaults are provided.

    Parameters
    ----------
    constructor_statement : str
        This is a string form of the function call used to create the object.  This is passed to the WaveformBase
        initializer as the parameter of the same name.  See the docstring for more information.
    data_functor : function
        Takes a 1-d array of time values and an array of (ell, m) values and returns the complex array of data.
    t : float array, optional
        Time values of the data.  Default is `np.linspace(-10., 100., num=1101))`.
    ell_min, ell_max : int, optional
        Smallest and largest ell value present in the data.  Defaults are 2 and 8.

    """
    t = np.array(kwargs.pop("t", np.linspace(-10.0, 100.0, num=1101)), dtype=float)
    frame = np.array(kwargs.pop("frame", []), dtype=np.quaternion)
    frameType = int(kwargs.pop("frameType", scri.Inertial))
    dataType = int(kwargs.pop("dataType", scri.h))
    r_is_scaled_out = bool(kwargs.pop("r_is_scaled_out", True))
    m_is_scaled_out = bool(kwargs.pop("m_is_scaled_out", True))
    ell_min = int(kwargs.pop("ell_min", abs(scri.SpinWeights[dataType])))
    ell_max = int(kwargs.pop("ell_max", 8))
    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")
    data = data_functor(t, sf.LM_range(ell_min, ell_max))
    w = scri.WaveformModes(
        t=t,
        frame=frame,
        data=data,
        history=["# Called from constant_waveform"],
        frameType=frameType,
        dataType=dataType,
        r_is_scaled_out=r_is_scaled_out,
        m_is_scaled_out=m_is_scaled_out,
        constructor_statement=constructor_statement,
        ell_min=ell_min,
        ell_max=ell_max,
    )
    return w


def constant_waveform(**kwargs):
    """WaveformModes object with constant values in each mode

    Additional keyword arguments are passed to `modes_constructor`.

    """
    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    def data_functor(t, LM):
        data = np.zeros((t.shape[0], LM.shape[0]), dtype=complex)
        for i, m in enumerate(LM[:, 1]):
            data[:, i] = m - 1j * m
        return data

    return modes_constructor(f"constant_waveform(**{kwargs})", data_functor)


def single_mode(ell, m, **kwargs):
    """WaveformModes object with 1 in selected slot and 0 elsewhere

    Additional keyword arguments are passed to `modes_constructor`.

    Parameters
    ----------
    ell, m : int
        The (ell, m) value of the nonzero mode

    """
    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    def data_functor(t, LM):
        data = np.zeros((t.shape[0], LM.shape[0]), dtype=complex)
        data[:, sf.LM_index(ell, m, min(LM[:, 0]))] = 1.0 + 0.0j
        return data

    return modes_constructor(f"single_mode({ell}, {m}, **{kwargs})", data_functor)


def random_waveform(**kwargs):
    """WaveformModes object filled with random data at each time step

    Additional keyword arguments are passed to `modes_constructor`.

    Parameters
    ----------
    uniform_time : bool, optional
        Use uniform, rather than random, time steps.  Default is False.
    begin, end : float, optional
        Initial and final times of the time data.  Default values are -10., 100.
    n_times : int, optional
        Number of time steps in the time data.  Default is 1101
    rotating : bool, optional
        If True, use a `Corotating` frame, with random orientation at each instant.  Default is True.

    """
    np.random.seed(hash("random_waveform") % 4294967294)
    begin = float(kwargs.pop("begin", -10.0))
    end = float(kwargs.pop("end", 100.0))
    n_times = int(kwargs.pop("n_times", 1101))
    if kwargs.pop("uniform_time", False):
        t = np.array(kwargs.pop("t", np.linspace(begin, end, num=n_times)), dtype=float)
    else:
        t = np.sort(np.random.uniform(begin, end, size=n_times))
    rotating = bool(kwargs.pop("rotating", True))

    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    def data_functor(tin, LM):
        data = np.random.normal(size=(tin.shape[0], LM.shape[0], 2)).view(complex)[:, :, 0]
        return data

    if rotating:
        frame = np.array([np.quaternion(*np.random.uniform(-1, 1, 4)).normalized() for t_i in t])
        return modes_constructor(
            f"random_waveform(**{kwargs})", data_functor, t=t, frame=frame, frameType=scri.Corotating
        )
    else:
        return modes_constructor(f"random_waveform(**{kwargs})", data_functor, t=t)


def random_waveform_proportional_to_time(**kwargs):
    """WaveformModes object filled with random data times the time coordinate

    Additional keyword arguments are passed to `modes_constructor`.

    Parameters
    ----------
    uniform_time : bool, optional
        Use uniform, rather than random, time steps.  Default is False.
    begin, end : float, optional
        Initial and final times of the time data.  Default values are -10., 100.
    n_times : int, optional
        Number of time steps in the time data.  Default is 1101
    rotating : bool, optional
        If True, use a `Corotating` frame, with random orientation at each instant.  Default is True.

    """
    np.random.seed(hash("random_waveform_proportional_to_time") % 4294967294)  # Use mod to get in an acceptable range
    begin = float(kwargs.pop("begin", -10.0))
    end = float(kwargs.pop("end", 100.0))
    n_times = int(kwargs.pop("n_times", 1101))
    if kwargs.pop("uniform_time", False):
        t = np.array(kwargs.pop("t", np.linspace(begin, end, num=n_times)), dtype=float)
    else:
        t = np.sort(np.random.uniform(begin, end, size=n_times))
    rotating = bool(kwargs.pop("rotating", True))

    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    def data_functor(tin, LM):
        return np.outer(tin, np.random.normal(size=(LM.shape[0], 2)).view(complex)[:, 0])

    if rotating:
        axis = np.quaternion(0.0, *np.random.uniform(-1, 1, size=3)).normalized()
        omega = 2 * np.pi * 4 / (t[-1] - t[0])
        frame = np.array([np.exp(axis * (omega * t_i / 2)) for t_i in t])
        return modes_constructor(
            f"random_waveform(**{kwargs})", data_functor, t=t, frame=frame, frameType=scri.Corotating
        )
    else:
        return modes_constructor(f"random_waveform(**{kwargs})", data_functor, t=t)


def single_mode_constant_rotation(**kwargs):
    """Return WaveformModes object a single nonzero mode, with phase proportional to time

    The waveform output by this function will have just one nonzero mode.  The behavior of that mode will be fairly
    simple; it will be given by exp(i*omega*t).  Note that omega can be complex, which gives damping.

    Parameters
    ----------
    s : int, optional
        Spin weight of the waveform field.  Default is -2.
    ell, m : int, optional
        The (ell, m) values of the nonzero mode in the returned waveform.  Default value is (abs(s), -abs(s)).
    ell_min, ell_max : int, optional
        Smallest and largest ell values present in the output.  Default values are abs(s) and 8.
    data_type : int, optional
        Default value is whichever psi_n corresponds to the input spin.  It is important to choose these, rather than
        `h` or `sigma` for the analytical solution to translations, which doesn't account for the direct contribution
        of supertranslations (as opposed to the indirect contribution, which involves moving points around).
    t_0, t_1 : float, optional
        Beginning and end of time.  Default values are -20. and 20.
    dt : float, optional
        Time step.  Default value is 0.1.
    omega : complex, optional
        Constant of proportionality such that nonzero mode is exp(i*omega*t).  Note that this can be complex, which
        implies damping.  Default is 0.5.

    """
    s = kwargs.pop("s", -2)
    ell = kwargs.pop("ell", abs(s))
    m = kwargs.pop("m", -ell)
    ell_min = kwargs.pop("ell_min", abs(s))
    ell_max = kwargs.pop("ell_max", 8)
    data_type = kwargs.pop("data_type", scri.DataType[scri.SpinWeights.index(s)])
    t_0 = kwargs.pop("t_0", -20.0)
    t_1 = kwargs.pop("t_1", 20.0)
    dt = kwargs.pop("dt", 1.0 / 10.0)
    t = np.arange(t_0, t_1 + dt, dt)
    n_times = t.size
    omega = complex(kwargs.pop("omega", 0.5))
    data = np.zeros((n_times, sf.LM_total_size(ell_min, ell_max)), dtype=complex)
    data[:, sf.LM_index(ell, m, ell_min)] = np.exp(1j * omega * t)

    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    return scri.WaveformModes(
        t=t,
        data=data,
        ell_min=ell_min,
        ell_max=ell_max,
        frameType=scri.Inertial,
        dataType=data_type,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )


def single_mode_proportional_to_time(**kwargs):
    """Return WaveformModes object a single nonzero mode, proportional to time

    The waveform output by this function will have just one nonzero mode.  The behavior of that mode will be
    particularly simple; it will just be proportional to time.

    Parameters
    ----------
    s : int, optional
        Spin weight of the waveform field.  Default is -2.
    ell, m : int, optional
        The (ell, m) values of the nonzero mode in the returned waveform.  Default value is (abs(s), -abs(s)).
    ell_min, ell_max : int, optional
        Smallest and largest ell values present in the output.  Default values are abs(s) and 8.
    data_type : int, optional
        Default value is whichever psi_n corresponds to the input spin.  It is important to choose these, rather than
        `h` or `sigma` for the analytical solution to translations, which doesn't account for the direct contribution
        of supertranslations (as opposed to the indirect contribution, which involves moving points around).
    t_0, t_1 : float, optional
        Beginning and end of time.  Default values are -20. and 20.
    dt : float, optional
        Time step.  Default value is 0.1.
    beta : complex, optional
        Constant of proportionality such that nonzero mode is beta*t.  Default is 1.

    """
    s = kwargs.pop("s", -2)
    ell = kwargs.pop("ell", abs(s))
    m = kwargs.pop("m", -ell)
    ell_min = kwargs.pop("ell_min", abs(s))
    ell_max = kwargs.pop("ell_max", 8)
    data_type = kwargs.pop("data_type", scri.DataType[scri.SpinWeights.index(s)])
    t_0 = kwargs.pop("t_0", -20.0)
    t_1 = kwargs.pop("t_1", 20.0)
    dt = kwargs.pop("dt", 1.0 / 10.0)
    t = np.arange(t_0, t_1 + dt, dt)
    n_times = t.size
    beta = kwargs.pop("beta", 1.0)
    data = np.zeros((n_times, sf.LM_total_size(ell_min, ell_max)), dtype=complex)
    data[:, sf.LM_index(ell, m, ell_min)] = beta * t

    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    return scri.WaveformModes(
        t=t,
        data=data,
        ell_min=ell_min,
        ell_max=ell_max,
        frameType=scri.Inertial,
        dataType=data_type,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )


def single_mode_proportional_to_time_supertranslated(**kwargs):
    """Return WaveformModes as in single_mode_proportional_to_time, with analytical supertranslation

    This function constructs the same basic object as the `single_mode_proportional_to_time`, but then applies an
    analytical supertranslation.  The arguments to this function are the same as to the other, with two additions:

    Additional parameters
    ---------------------
    supertranslation : complex array, optional
        Spherical-harmonic modes of the supertranslation to apply to the waveform.  This is overwritten by
         `space_translation` if present.  Default value is `None`.
    space_translation : float array of length 3, optional
        This is just the 3-vector representing the displacement to apply to the waveform.  Note that if
        `supertranslation`, this parameter overwrites it.  Default value is [1.0, 0.0, 0.0].

    """
    s = kwargs.pop("s", -2)
    ell = kwargs.pop("ell", abs(s))
    m = kwargs.pop("m", -ell)
    ell_min = kwargs.pop("ell_min", abs(s))
    ell_max = kwargs.pop("ell_max", 8)
    data_type = kwargs.pop("data_type", scri.DataType[scri.SpinWeights.index(s)])
    t_0 = kwargs.pop("t_0", -20.0)
    t_1 = kwargs.pop("t_1", 20.0)
    dt = kwargs.pop("dt", 1.0 / 10.0)
    t = np.arange(t_0, t_1 + dt, dt)
    n_times = t.size
    beta = kwargs.pop("beta", 1.0)
    data = np.zeros((n_times, sf.LM_total_size(ell_min, ell_max)), dtype=complex)
    data[:, sf.LM_index(ell, m, ell_min)] = beta * t
    supertranslation = np.array(kwargs.pop("supertranslation", np.array([], dtype=complex)), dtype=complex)
    if "space_translation" in kwargs:
        if supertranslation.size < 4:
            supertranslation.resize((4,))
        supertranslation[1:4] = -sf.vector_as_ell_1_modes(kwargs.pop("space_translation"))
    supertranslation_ell_max = int(math.sqrt(supertranslation.size) - 1)
    if supertranslation_ell_max * (supertranslation_ell_max + 2) + 1 != supertranslation.size:
        raise ValueError(f"Bad number of elements in supertranslation: {supertranslation.size}")
    for i, (ellpp, mpp) in enumerate(sf.LM_range(0, supertranslation_ell_max)):
        if supertranslation[i] != 0.0:
            mp = m + mpp
            for ellp in range(ell_min, min(ell_max, (ell + ellpp)) + 1):
                if ellp >= abs(mp):
                    addition = (
                        beta
                        * supertranslation[i]
                        * math.sqrt(((2 * ellpp + 1) * (2 * ell + 1) * (2 * ellp + 1)) / (4 * math.pi))
                        * sf.Wigner3j(ellpp, ell, ellp, 0, -s, s)
                        * sf.Wigner3j(ellpp, ell, ellp, mpp, m, -mp)
                    )
                    if (s + mp) % 2 == 1:
                        addition *= -1
                    data[:, sf.LM_index(ellp, mp, ell_min)] += addition

    if kwargs:
        import pprint

        warnings.warn(f"\nUnused kwargs passed to this function:\n{pprint.pformat(kwargs, width=1)}")

    return scri.WaveformModes(
        t=t,
        data=data,
        ell_min=ell_min,
        ell_max=ell_max,
        frameType=scri.Inertial,
        dataType=data_type,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )


def fake_precessing_waveform(
    t_0=-20.0,
    t_1=20_000.0,
    dt=0.1,
    ell_max=8,
    mass_ratio=2.0,
    precession_opening_angle=np.pi / 6.0,
    precession_opening_angle_dot=None,
    precession_relative_rate=0.1,
    precession_nutation_angle=None,
    inertial=True,
):
    """Construct a strain waveform with realistic precession effects.

    This model is intended to be weird enough that it breaks any overly simplistic assumptions about
    waveform symmetries while still being (mostly) realistic.

    This waveform uses only the very lowest-order terms from PN theory to evolve the orbital
    frequency up to a typical constant value (with a smooth transition), and to construct modes that
    have very roughly the correct amplitudes as a function of the orbital frequency.  Modes with
    equal ell values but opposite m values are modulated antisymmetrically, though this modulation
    decays quickly after merger -- roughly as it would behave in a precessing system.  The modes are
    then smoothly transitioned to an exponential decay after merger.  The frame is a simulated
    precession involving the basic orbital rotation precessing about a cone of increasing opening
    angle and nutating about that cone on the orbital time scale, but settling down to a constant
    direction shortly after merger.  (Though there is little precise physical content, these
    features are all found in real waveforms.)  If the input argument `inertial` is `True` (the
    default), the waveform is transformed back to the inertial frame before returning.

    Parameters
    ==========
    t_0: float [defaults to -20.0]
    t_1: float [defaults to 20_000.0]
        The initial and final times in the output waveform.  Note that the merger is placed 100.0
        time units before `t_1`, and several transitions are made before this, so `t_0` must be that
        far ahead of `t_1`.
    dt: float [defaults to 0.1]
        Spacing of output time series.
    ell_max: int [defaults to 8]
        Largest ell value in the output modes.
    mass_ratio: float [defaults to 2.0]
        Ratio of BH masses to use as input to rough approximations for orbital evolution and mode
        amplitudes.
    precession_opening_angle: float [defaults to pi/6]
        Opening angle of the precession cone.
    precession_opening_angle_dot: float [defaults to 2*precession_opening_angle/(t_merger-t_0)]
        Rate at which precession cone opens.
    precession_relative_rate: float [defaults to 0.1]
        Fraction of the magnitude of the orbital angular velocity at which it precesses.
    precession_nutation_angle: float [defaults to precession_opening_angle/10]
        Angle (relative to precession_opening_angle) by which the orbital angular velocity nutates.

    """
    import warnings
    import numpy as np
    import quaternion
    from quaternion.calculus import indefinite_integral
    from .utilities import transition_function, transition_to_constant

    if mass_ratio < 1.0:
        mass_ratio = 1.0 / mass_ratio

    s = -2
    ell_min = abs(s)
    data_type = scri.h

    nu = mass_ratio / (1 + mass_ratio) ** 2
    t = np.arange(t_0, t_1 + 0.99 * dt, dt)
    t_merger = t_1 - 100.0
    i_merger = np.argmin(abs(t - t_merger))
    if i_merger < 20:
        raise ValueError(f"Insufficient space between initial time (t={t_merger}) and merger (t={t_0}).")
    n_times = t.size
    data = np.zeros((n_times, sf.LM_total_size(ell_min, ell_max)), dtype=complex)

    # Get a rough approximation to the phasing through merger
    tau = nu * (t_merger - t) / 5
    with warnings.catch_warnings():  # phi and omega will have NaNs after t_merger for now
        warnings.simplefilter("ignore")
        phi = -4 * tau ** (5 / 8)
        omega = (nu / 2) * tau ** (-3 / 8)

    # Now, transition omega smoothly up to a constant value of 0.25
    omega_transition_width = 5.0
    i1 = np.argmin(np.abs(omega[~np.isnan(omega)] - 0.25))
    i0 = np.argmin(np.abs(t - (t[i1] - omega_transition_width)))
    transition = transition_function(t, t[i0], t[i1])
    zero_point_two_five = 0.25 * np.ones_like(t)
    omega[:i1] = omega[:i1] * (1 - transition[:i1]) + zero_point_two_five[:i1] * transition[:i1]
    omega[i1:] = 0.25

    # Integrate phi after i0 to agree with the new omega
    phi[i0:] = phi[i0] + indefinite_integral(omega[i0:], t[i0:])

    # Construct ringdown-transition function
    ringdown_transition_width = 20
    t0 = t_merger
    i0 = np.argmin(np.abs(t - t_merger))
    i1 = np.argmin(np.abs(t - (t[i0] + ringdown_transition_width)))
    t0 = t[i0]
    t1 = t[i1]
    transition = transition_function(t, t0, t1)
    ringdown = np.ones_like(t)
    ringdown[i0:] = ringdown[i0:] * (1 - transition[i0:]) + 2.25 * np.exp(-(t[i0:] - t_merger) / 11.5) * transition[i0:]

    # Construct frame
    if precession_opening_angle_dot is None:
        precession_opening_angle_dot = 2.0 * precession_opening_angle / (t[i1] - t[0])
    if precession_nutation_angle is None:
        precession_nutation_angle = precession_opening_angle / 10.0
    R_orbital = np.exp(phi * quaternion.z / 2)
    R_opening = np.exp(
        transition_to_constant(precession_opening_angle + precession_opening_angle_dot * t, t, t0, t1)
        * quaternion.x
        / 2
    )
    R_precession = np.exp(transition_to_constant(phi / precession_relative_rate, t, t0, t1) * quaternion.z / 2)
    R_nutation = np.exp(precession_nutation_angle * transition * quaternion.x / 2)
    frame = (
        R_orbital * R_nutation * R_orbital.conjugate() * R_precession * R_opening * R_precession.conjugate() * R_orbital
    )
    frame = frame[0].sqrt().conjugate() * frame  # Just give the initial angle a weird little tweak to screw things up

    # Construct the modes
    x = omega ** (2 / 3)
    modulation = transition_function(t, t[i0], t[i1], 1, 0) * np.cos(phi) / 40.0
    for ell in range(ell_min, ell_max + 1):
        for m in range(-ell, ell + 1):
            data[:, sf.LM_index(ell, m, ell_min)] = pn_leading_order_amplitude(ell, m, x, mass_ratio=mass_ratio) * (
                1 + np.sign(m) * modulation
            )

    # Apply ringdown (mode amplitudes are constant after t_merger)
    data *= ringdown[:, np.newaxis]

    h_corot = scri.WaveformModes(
        t=t,
        frame=frame,
        data=data,
        ell_min=ell_min,
        ell_max=ell_max,
        frameType=scri.Corotating,
        dataType=data_type,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )

    if inertial:
        return h_corot.to_inertial_frame()
    else:
        return h_corot


def pn_leading_order_amplitude(ell, m, x, mass_ratio=1.0):
    """Return the leading-order amplitude of r*h/M in PN theory

    These expressions are from Eqs. (330) of Blanchet's Living Review (2014).

    Note that `x` is just the orbital angular velocity to the (2/3) power.

    """
    from scipy.special import factorial, factorial2

    if m < 0:
        return (-1) ** ell * np.conjugate(pn_leading_order_amplitude(ell, -m, x, mass_ratio=mass_ratio))

    if mass_ratio < 1.0:
        mass_ratio = 1.0 / mass_ratio
    nu = mass_ratio / (1 + mass_ratio) ** 2
    X1 = mass_ratio / (mass_ratio + 1)
    X2 = 1 / (mass_ratio + 1)

    def sigma(ell):
        return X2 ** (ell - 1) + (-1) ** ell * X1 ** (ell - 1)

    if (ell + m) % 2 == 0:
        amplitude = (
            (
                (-1) ** ((ell - m + 2) / 2)
                / (2 ** (ell + 1) * factorial((ell + m) // 2) * factorial((ell - m) // 2) * factorial2(2 * ell - 1))
            )
            * np.sqrt(
                (5 * (ell + 1) * (ell + 2) * factorial(ell + m) * factorial(ell - m))
                / (ell * (ell - 1) * (2 * ell + 1))
            )
            * sigma(ell)
            * (1j * m) ** ell
            * x ** (ell / 2 - 1)
        )
    else:
        amplitude = (
            (
                (-1) ** ((ell - m - 1) / 2)
                / (
                    2 ** (ell - 1)
                    * factorial((ell + m - 1) // 2)
                    * factorial((ell - m - 1) // 2)
                    * factorial2(2 * ell + 1)
                )
            )
            * np.sqrt(
                (5 * (ell + 2) * (2 * ell + 1) * factorial(ell + m) * factorial(ell - m))
                / (ell * (ell - 1) * (ell + 1))
            )
            * sigma(ell + 1)
            * 1j
            * (1j * m) ** ell
            * x ** ((ell - 1) / 2)
        )

    return 8 * np.sqrt(np.pi / 5) * nu * x * amplitude


def create_fake_finite_radius_strain_h5file(
    output_file_path="./rh_FiniteRadii_CodeUnits.h5",
    n_subleading=3,
    amp=1.0,
    t_0=0.0,
    t_1=3000.0,
    dt=0.1,
    r_min=100.0,
    r_max=600.0,
    n_radii=24,
    ell_max=8,
    initial_adm_energy=0.99,
    avg_lapse=0.99,
    avg_areal_radius_diff=1.0,
    mass_ratio=1.0,
    precession_opening_angle=0.0,
    **kwargs,
):
    """
    Create an HDF5 file with fake finite-radius GW strain data in the NRAR format, as
    used by the Spectral Einstein Code (SpEC). The asymptotic waveform is created by
    the function scri.sample_waveforms.fake_precessing_waveform and then radius-dependent
    terms are added to it. These subleading artificial "near-zone" effects are simple
    sinusoids. The finite-radius waveform, scaled by a factor of the radius, is then,

        r*h(r) = h_0 + h_1*r**-1 + h_2*r**-2 + ... + h_n*r**-n

    where h_0 is the waveform from scri.sample_waveforms.fake_precessing_waveform() and n is
    chosen by the user.

    Finally, the finite-radius waveforms as output by SpEC uses simulation coordinates for
    radius and time, which do not perfectly parametrize outgoing null rays. This function
    approximates these simulation coordinates so that the data in the resulting HDF5 file
    most nearly mimics that of a SpEC HDF5 output file.

    Parameters
    ==========
    output_file_path: str
        The name and path of the output file. The filename must be "rh_FiniteRadius_CodeUnits.h5"
        if you wish to be able to run scri.extrapolation.extrapolate on it.
    n_subleading: int [defaults to 3]
        The number of subleading, radius-dependent terms to add to the asymptotic waveform.
    amp: float [defaults to 1.0]
        The amplitude of the subleading, radius-dependent terms.
    t_0: float [defaults to 0.0]
    t_1: float [defaults to 3000.0]
        The initial and final times in the output waveform. Note that the merger is placed
        100.0 time units before `t_1`, and several transitions are made before this, so `t_0`
        must be that far ahead of `t_1`.
    dt: float [defaults to 0.1]
        Spacing of output time series.
    r_min: float [defaults to 100.0]
    r_max: float [defaults to 600.0]
    n_radii: float [defaults to 24]
        The minimum and maximum radius, and the number of radii between these values, at which
        to produce a finite-radius waveform. These will be equally spaced in inverse radius.
    ell_max: int [defaults to 8]
        Largest ell value in the output modes.
    initial_adm_energy: float [defaults to 0.99]
        The intial ADM energy as would be computed by the SpEC intitial data solver.
    avg_lapse: float [defaults to 0.99]
        The value of the lapse averaged over a sphere at
    avg_areal_radius_diff: float [defaults to 1.0]
        How much on average the areal radius is larger than the coord radius, may be negative.
    mass_ratio: float [defaults to 1.0]
        Ratio of BH masses to use as input to rough approximations for orbital evolution
        and mode amplitudes.
    precession_opening_angle: float [defaults to 0.0]
        Opening angle of the precession cone. The default results in no precession. If
        this value is non-zero, then the following options from fake_precessing_waveform
        may be supplied as kwagrs:
            * precession_opening_angle_dot
            * precession_relative_rate
            * precession_nutation_angle
        See the help text of fake_precessing_waveform for documentation.

    """
    import h5py
    from scipy.interpolate import CubicSpline

    with h5py.File(output_file_path, "x") as h5file:

        # Set up an H5 group for each radius
        coord_radii = (1 / np.linspace(1 / r_min, 1 / r_max, n_radii)).astype(int)
        groups = [f"R{R:04}.dir" for R in coord_radii]
        for group in groups:
            h5file.create_group(group)

        # Create version history dataset
        # We mimic how SpEC encodes the strings in VersionHist.ver to ensure we
        # don't hit any errors due to the encoding.
        version_hist = [["ef51849550a1d8a5bbdd810c7debf0dd839e86dd", "Overall sign change in complex strain h."]]
        reencoded_version_hist = [
            [entry[0].encode("ascii", "ignore"), entry[1].encode("ascii", "ignore")] for entry in version_hist
        ]
        dset = h5file.create_dataset(
            "VersionHist.ver",
            (len(version_hist), 2),
            maxshape=(None, 2),
            data=reencoded_version_hist,
            dtype=h5py.special_dtype(vlen=bytes),
        )
        dset.attrs.create("Version", len(version_hist), shape=(1,), dtype=np.uint64)

        # Generate the asymptotic waveform
        h0 = scri.sample_waveforms.fake_precessing_waveform(
            t_0=t_0, t_1=t_1, dt=dt, ell_max=ell_max, precession_opening_angle=precession_opening_angle, **kwargs
        )
        n_times = int(h0.t.shape[0] + r_max / dt)
        t_1 += r_max
        all_times = np.linspace(0, t_1, n_times)

        # Set auxiliary datasetsss
        for i in range(len(groups)):
            # Set coordinate radius dataset
            coord_radius = np.vstack((all_times, [coord_radii[i]] * n_times)).T
            dset = h5file.create_dataset(f"{groups[i]}/CoordRadius.dat", data=coord_radius)
            dset.attrs.create("Legend", ["time", "CoordRadius"])

            # Set areal radius dataset
            areal_radius = coord_radius
            areal_radius[:, 1] += avg_areal_radius_diff
            dset = h5file.create_dataset(f"{groups[i]}/ArealRadius.dat", data=areal_radius)
            dset.attrs.create("Legend", ["time", "ArealRadius"])

            # Set initial ADM energy dataset
            seg_start_times = h0.t[:: int(n_times / 20)]
            adm_energy_dset = np.vstack((seg_start_times, [initial_adm_energy] * len(seg_start_times))).T
            dset = h5file.create_dataset(f"{groups[i]}/InitialAdmEnergy.dat", data=adm_energy_dset)
            dset.attrs.create("Legend", ["time", "InitialAdmEnergy"])

            # Set average lapse dataset
            avg_lapse_dset = np.vstack((all_times, [avg_lapse] * n_times)).T
            dset = h5file.create_dataset(f"{groups[i]}/AverageLapse.dat", data=avg_lapse_dset)
            dset.attrs.create("Legend", ["time", "AverageLapse"])

            # Set finite radius data
            R = areal_radius[0, 1]
            tortoise_coord = R + 2 * initial_adm_energy * np.log(R / (2 * initial_adm_energy) - 1)
            # Compute the approximate time in SpEC simulation coordinates
            simulation_time = (h0.t + tortoise_coord) * np.sqrt(1 - 2 * initial_adm_energy / R) / avg_lapse
            for l in range(2, ell_max + 1):
                for m in range(-l, l + 1):
                    index = sf.LM_index(l, m, 2)
                    new_data = h0.data[:, index]
                    for n in range(1, n_subleading + 1):
                        new_data += amp * R ** -n * np.exp((1j * n * 50 * np.pi / h0.n_times) * h0.t) * h0.abs[:, index]
                    new_data = CubicSpline(simulation_time, new_data)(all_times)
                    new_data[(all_times < simulation_time[0]) | (all_times > simulation_time[-1])] = 0.0
                    new_data += (1 + 1j) * 1e-14 * all_times
                    new_dset = np.vstack((all_times, new_data.real, new_data.imag)).T
                    dset = h5file.create_dataset(f"{groups[i]}/Y_l{l}_m{m}.dat", data=new_dset)
                    dset.attrs.create(
                        "Legend",
                        [
                            "time",
                            f"Re[rh]_l{l}_m{m}(R={coord_radii[i]}.00)",
                            f"Im[rh]_l{l}_m{m}(R={coord_radii[i]}.00)",
                        ],
                    )



# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import functools
import numpy as np
import numba
from spherical_functions import clebsch_gordan as CG
from . import jit


def swsh_indices_to_matrix_indices(matrix_iterator):
    """Convert SWSH-indexed function into caching sparse matrix-indexed function

    This function is designed to decorate functions with signature (ell_min,
    ell_max) that yield tuples of (ellp, mp, ell, m, value), and convert them into
    functions with the same signature but returning tuples of (rows, columns,
    values).  Here, the rows are row indices of a matrix, corresponding to the
    (ellp, mp) in standard `spherical_functions` order.  Similarly the columns are
    column indices corresponding to (ell, m).  The result can be passed into the
    `sparse_expectation_value` function.

    """
    from spherical_functions import LM_index

    @functools.lru_cache()  # Decorate the decorated function with caching
    def wrapper(ell_min, ell_max, *args, **kwargs):
        rows, columns, values = zip(
            *(  # zip* is the inverse of zip
                (LM_index(ellp, mp, ell_min), LM_index(ell, m, ell_min), value)
                for ellp, mp, ell, m, value in matrix_iterator(ell_min, ell_max, *args, **kwargs)
            )
        )
        return np.array(rows, dtype=int), np.array(columns, dtype=int), np.array(values)

    functools.update_wrapper(wrapper, matrix_iterator)  # Copy over the name, docstring, and such

    return wrapper


@jit
def sparse_expectation_value(abar, rows, columns, values, b):
    """Helper function for the main calculation of `matrix_expectation_value`

    Computes ⟨a|M|b⟩, assuming that ⟨a| and |b⟩ have the same shapes, and that
    (rows, columns, values) are corresponding arrays containing the nonzero
    elements of the matrix M.

    Parameters
    ----------
    abar : ndarray
        The data for ⟨a| (after complex conjugation was applied).

    rows, columns : list
        Lists of same length, containing indices into (respectively)
        abar and b.

    values : ndarray
        Should have shape (n,) where n is the same length as rows and
        columns.  These values are the matrix elements themselves to
        be summed.

    b : ndarray
        The data for |b⟩. Must have the same shape as abar.

    Returns
    -------
    expectation_value : ndarray

    """
    n_times = abar.shape[0]
    n_elements = rows.shape[0]
    expectation_value = np.zeros(n_times, dtype=numba.complex128)
    for i_time in range(n_times):
        for i_element in range(n_elements):
            expectation_value[i_time] += (
                abar[i_time, rows[i_element]] * b[i_time, columns[i_element]] * values[i_element]
            )
    return expectation_value


def matrix_expectation_value(a, M, b, allow_LM_differ=False, allow_times_differ=False):
    """The matrix expectation value ⟨a|M|b⟩(u), where M is a linear operator

    Treat two spin-s waveforms a, b (in the modal representation) as vectors.
    Then we can form the 'expectation value'

      ⟨a|M|b⟩(u) ≡ ∑ⱼₙₗₘ āⱼₙ(u) Mⱼₙₗₘ bₗₘ(u).

    For efficiency with sparse matrices, M should be implemented as a generator
    which yields tuples (j, n, l, m, Mⱼₙₗₘ) for only the non-vanishing matrix
    elements.

    Parameters
    ----------
    a : WaveformModes object
        The waveform |a⟩. This function is "antilinear" in `a`.

    M : callable
        M will be called like
        `for ellp, mp, ell, m, M_el in M(ell_min, ell_max):`
        This is best implemented as a generator yielding tuples
        `(ellp, mp, ell, m, M_{ellp mp ell m})`
        which range over ell, ellp values up to and including ell_max,
        for only the non-vanishing matrix elements of M.

    b : WaveformModes object
        The waveform |b⟩. This function is linear in `b`.

    allow_LM_differ : bool, optional [default: False]
        If True and if the set of (ell,m) modes between a and b
        differ, then the inner product will be computed using the
        intersection of the set of modes.

    allow_times_differ: bool, optional [default: False]
        If True and if the set of times between a and b differ,
        then both WaveformModes will be interpolated to the
        intersection of the set of times.

    Returns
    -------
    times, expect_val : ndarray, complex ndarray
        Resulting tuple has length two.  The first element is the set
        of times u for the timeseries.  The second element is the
        timeseries of ⟨a|M|b⟩(u).

    """
    from .extrapolation import intersection
    from spherical_functions import LM_index

    if a.spin_weight != b.spin_weight:
        raise ValueError("Spin weights must match in matrix_expectation_value")

    LM_clip = slice(a.ell_min, a.ell_max + 1)
    if (a.ell_min != b.ell_min) or (a.ell_max != b.ell_max):
        if allow_LM_differ:
            LM_clip = slice(max(a.ell_min, b.ell_min), min(a.ell_max, b.ell_max) + 1)
            if LM_clip.start >= LM_clip.stop:
                raise ValueError("Intersection of (ell,m) modes is " "empty.  Assuming this is not desired.")
        else:
            raise ValueError(
                "ell_min and ell_max must match in matrix_expectation_value " "(use allow_LM_differ=True to override)"
            )

    t_clip = None
    if not np.array_equal(a.t, b.t):
        if allow_times_differ:
            t_clip = intersection(a.t, b.t)
        else:
            raise ValueError(
                "Time samples must match in matrix_expectation_value " "(use allow_times_differ=True to override)"
            )

    ##########

    times = a.t
    A = a
    B = b

    if LM_clip is not None:
        A = A[:, LM_clip]
        B = B[:, LM_clip]

    if t_clip is not None:
        times = t_clip
        A = A.interpolate(t_clip)
        B = B.interpolate(t_clip)

    ##########
    ## Time for actual math!

    Abar_data = np.conj(A.data)
    B_data = B.data

    ell_min = LM_clip.start
    ell_max = LM_clip.stop - 1

    rows, columns, values = M(ell_min, ell_max)

    return (times, sparse_expectation_value(Abar_data, rows, columns, values, B_data))


def energy_flux(h):
    """Compute energy flux from waveform

    This implements Eq. (2.8) from Ruiz et al. (2008) [0707.4654].

    """
    from .waveform_modes import WaveformModes
    from . import h as htype
    from . import hdot as hdottype

    if not isinstance(h, WaveformModes):
        raise ValueError(
            f"Energy flux can only be calculated from a `WaveformModes` object; this object is of type `{type(h)}`."
        )
    if h.dataType == hdottype:
        hdot = h.data
    elif h.dataType == htype:
        hdot = h.data_dot
    else:
        raise ValueError(
            f"Input argument is expected to have data of type `h` or `hdot`; this waveform data has type `{h.data_type_string}`"
        )

    # No need to use matrix_expectation_value here
    Edot = np.einsum("ij, ij -> i", hdot.conjugate(), hdot).real

    Edot /= 16.0 * np.pi

    return Edot


@swsh_indices_to_matrix_indices
def p_z(ell_min, ell_max, s=-2):
    """Generator for pᶻ matrix elements (for use with `matrix_expectation_value`)

    This function is specific to the case where waveforms have s=-2.

      pᶻ = cosθ = 2 √(π/3) Y₁₀

    This is what Ruiz et al. (2008) [0707.4654] call "lᶻ".

    The matrix elements yielded are

      ⟨s,j,n|cosθ|s,l,m⟩ = √[(2l+1)/(2j+1)] ⟨l,m,1,0|j,m⟩ ⟨l,-s,1,0|j,-s⟩

    where the terms on the last line are the ordinary Clebsch-Gordan coefficients.
    Because of the magnetic selection rules, we only have nonzero elements for
    n==m.

    We could have used `_swsh_Y_mat_el` but I am just preemptively combining the
    prefactors.

    """

    for ell in range(ell_min, ell_max + 1):
        ellp_min = max(ell_min, ell - 1)
        ellp_max = min(ell_max, ell + 1)
        for ellp in range(ellp_min, ellp_max + 1):
            for m in range(-ell, ell + 1):
                if (m < -ellp) or (m > ellp):
                    continue
                cg1 = CG(ell, m, 1, 0, ellp, m)
                cg2 = CG(ell, -s, 1, 0, ellp, -s)
                prefac = np.sqrt((2.0 * ell + 1.0) / (2.0 * ellp + 1.0))
                yield ellp, m, ell, m, (prefac * cg1 * cg2)


@swsh_indices_to_matrix_indices
def p_plusminus(ell_min, ell_max, sign, s=-2):
    """Produce the function p_plus or p_minus, based on sign.

      p⁺ = -√(8π/3) Y₁₊₁ = sinθ exp[+iϕ]
      p⁻ = +√(8π/3) Y₁₋₁ = sinθ exp[-iϕ]

    These are what Ruiz et al. (2008) [0707.4654] call "l⁺" and "l⁻".

    We use `swsh_Y_mat_el` to compute the matrix elements.  Notice that since the
    operator has definite m=±1, we only have mp==m±1 nonvanishing in the
    matrix elements.

    """

    if (sign != 1) and (sign != -1):
        raise ValueError("sign must be either 1 or -1 in j_plusminus")

    prefac = -1.0 * sign * np.sqrt(8.0 * np.pi / 3.0)

    def swsh_Y_mat_el(s, l3, m3, l1, m1, l2, m2):
        """Compute a matrix element treating Yₗₘ as a linear operator

        From the rules for the Wigner D matrices, we get the result that

          ⟨s,l₃,m₃|Yₗ₁ₘ₁|s,l₂,m₂⟩ =
            √[(2l₁+1)(2l₂+1)/(4π(2l₃+1))] ⟨l₁,m₁,l₂,m₂|l₃,m₃⟩ ⟨l₁,0,l₂,−s|l₃,−s⟩

        where the terms on the last line are the ordinary Clebsch-Gordan
        coefficients.  See, e.g., Campbell and Morgan (1971).

        """    	
        cg1 = CG(l1, m1, l2, m2, l3, m3)
        cg2 = CG(l1, 0.0, l2, -s, l3, -s)

        return np.sqrt((2.0 * l1 + 1.0) * (2.0 * l2 + 1.0) / (4.0 * np.pi * (2.0 * l3 + 1))) * cg1 * cg2

    for ell in range(ell_min, ell_max + 1):
        ellp_min = max(ell_min, ell - 1)
        ellp_max = min(ell_max, ell + 1)
        for ellp in range(ellp_min, ellp_max + 1):
            for m in range(-ell, ell + 1):
                mp = round(m + 1 * sign)
                if (mp < -ellp) or (mp > ellp):
                    continue
                yield (ellp, mp, ell, m, (prefac * swsh_Y_mat_el(s, ellp, mp, 1.0, sign, ell, m)))


p_plus = functools.partial(p_plusminus, sign=+1)
p_minus = functools.partial(p_plusminus, sign=-1)
p_plus.__doc__ = p_plusminus.__doc__
p_minus.__doc__ = p_plusminus.__doc__


def momentum_flux(h):
    """Compute momentum flux from waveform

    This implements Eq. (2.11) from Ruiz et al. (2008) [0707.4654] by using
    `matrix_expectation_value` with `p_z`, `p_plus`, and `p_minus`.

    """
    from .waveform_modes import WaveformModes
    from . import h as htype
    from . import hdot as hdottype

    if not isinstance(h, WaveformModes):
        raise ValueError(
            f"Momentum flux can only be calculated from a `WaveformModes` object; this object is of type `{type(h)}`."
        )
    if h.dataType == hdottype:
        hdot = h
    elif h.dataType == htype:
        hdot = h.copy()
        hdot.dataType = hdottype
        hdot.data = h.data_dot
    else:
        raise ValueError(
            f"Input argument is expected to have data of type `h` or `hdot`; this waveform data has type `{h.data_type_string}`"
        )

    pdot = np.zeros((hdot.n_times, 3), dtype=float)

    _, p_plus_dot = matrix_expectation_value(hdot, functools.partial(p_plus, s=-2), hdot)
    _, p_minus_dot = matrix_expectation_value(hdot, functools.partial(p_minus, s=-2), hdot)
    _, p_z_dot = matrix_expectation_value(hdot, functools.partial(p_z, s=-2), hdot)

    # Convert into (x,y,z) basis
    pdot[:, 0] = 0.5 * (p_plus_dot.real + p_minus_dot.real)
    pdot[:, 1] = 0.5 * (p_plus_dot.imag - p_minus_dot.imag)
    pdot[:, 2] = p_z_dot.real

    pdot /= 16.0 * np.pi

    return pdot


@swsh_indices_to_matrix_indices
def j_z(ell_min, ell_max):
    r"""Generator for jᶻ matrix elements (for use with matrix_expectation_value)

    Matrix elements yielded are

      ⟨j,n|jᶻ|l,m⟩ = i m δⱼₗ δₙₘ

    This follows the convention for jᶻ from Ruiz et al. (2008) [0707.4654]

    """
    for ell in range(ell_min, ell_max + 1):
        for m in range(-ell, ell + 1):
            yield ell, m, ell, m, (1.0j * m)


@swsh_indices_to_matrix_indices
def j_plusminus(ell_min, ell_max, sign):
    r"""Produce the function j_plus or j_minus, based on sign.

    The conventions for these matrix elements, to agree with Ruiz et al. (2008)
    [0707.4654], should be:

      ⟨j,n|j⁺|l,m⟩ = i √[(l-m)(l+m+1)] δⱼₗ δₙₘ₊₁
      ⟨j,n|j⁻|l,m⟩ = i √[(l+m)(l-m+1)] δⱼₗ δₙₘ₋₁

    The spinsfast function `ladder_operator_coefficient(l, m)` gives
    √[(l-m)(l+m+1)].

    """
    from spherical_functions import ladder_operator_coefficient as ladder

    if (sign != 1) and (sign != -1):
        raise ValueError("sign must be either 1 or -1 in j_plusminus")

    for ell in range(ell_min, ell_max + 1):
        for m in range(-ell, ell + 1):
            mp = round(m + 1 * sign)
            if (mp < -ell) or (mp > ell):
                continue
            yield ell, mp, ell, m, (1.0j * ladder(ell, m * sign))


j_plus = functools.partial(j_plusminus, sign=+1)
j_minus = functools.partial(j_plusminus, sign=-1)
j_plus.__doc__ = j_plusminus.__doc__
j_minus.__doc__ = j_plusminus.__doc__


def angular_momentum_flux(h, hdot=None):
    """Compute angular momentum flux from waveform

    This implements Eq. (2.24) from Ruiz et al. (2008) [0707.4654] by using
    `matrix_expectation_value` with (internal) helper functions `j_z`, `j_plus`, and
    `j_minus`.

    """
    from .waveform_modes import WaveformModes
    from . import h as htype
    from . import hdot as hdottype

    if not isinstance(h, WaveformModes):
        raise ValueError(
            f"Angular momentum flux can only be calculated from a `WaveformModes` object; `h` is of type `{type(h)}`."
        )
    if (hdot is not None) and (not isinstance(hdot, WaveformModes)):
        raise ValueError(
            f"Angular momentum flux can only be calculated from a `WaveformModes` object; `hdot` is of type `{type(hdot)}`."
        )
    if h.dataType == htype:
        if hdot is None:
            hdot = h.copy()
            hdot.dataType = hdottype
            hdot.data = h.data_dot
        elif hdot.dataType != hdottype:
            raise ValueError(
                f"Input argument `hdot` is expected to have data of type `hdot`; this `hdot` waveform data has type `{h.data_type_string}`"
            )
    else:
        raise ValueError(
            f"Input argument `h` is expected to have data of type `h`; this `h` waveform data has type `{h.data_type_string}`"
        )

    jdot = np.zeros((hdot.n_times, 3), dtype=float)

    _, j_plus_dot = matrix_expectation_value(hdot, j_plus, h)
    _, j_minus_dot = matrix_expectation_value(hdot, j_minus, h)
    _, j_z_dot = matrix_expectation_value(hdot, j_z, h)

    # Convert into (x,y,z) basis
    jdot[:, 0] = 0.5 * (j_plus_dot.real + j_minus_dot.real)
    jdot[:, 1] = 0.5 * (j_plus_dot.imag - j_minus_dot.imag)
    jdot[:, 2] = j_z_dot.real

    jdot /= -16.0 * np.pi

    return jdot


def boost_flux(h, hdot=None):
    """Computes the boost flux from the waveform.

    This implements Eq. (C.1) from Flanagan & Nichols (2016) [1510.03386] for the
    boost vector field.

    Notes
    -----
    Boost flux is analytically calculated using the Wald & Zoupas formalism,
    following the calculation by Flanagan & Nichols [1510.03386].  Then, the NP
    formalism is implemented to make it computationally feasible.  The conventions
    of Ruiz et al. (2008) [0707.4654] are followed.

    The expression for boost flux is

      boost_flux(h) = (-1/32π) (
        (1/8) [⟨ð̄N|χ|ð̄h⟩ - ⟨ðN|χ|ðh⟩ + ⟨ð̄h|χ|ð̄N⟩ - ⟨ðh|χ|ðN⟩ + 6⟨N|χ|h⟩ + 6⟨h|χ|N⟩]
        - (1/4) [⟨ðN|ðχ|h⟩ + ⟨h|ð̄χ|ðN⟩]
        - (u/2) ⟨N|χ|N⟩
      )

    h and N have spin weight s=-2.  χ is an l=1, s=0 function characterizing the
    direction of the boost.  ð and ð̄ are spin raising and lowering operators.

    The matrix elements for most of the terms are identical to pᶻ, p⁺, and p⁻
    matrix elements, except that input for s is different.  For example, ⟨ð̄N|χ|ð̄h⟩
    will have s = -3.  There are two terms ⟨ðN|ðχ|h⟩ and ⟨h|ð̄χ|ðN⟩ for which matrix
    elements have been defined below.

    """
    from .waveform_modes import WaveformModes
    from . import h as htype
    from . import hdot as hdottype

    # We start by defining matrix elements that are not present already in this code.

    @swsh_indices_to_matrix_indices
    def eth_chi_z(ell_min, ell_max, s=-2):
        """Generator for the matrix element ⟨ðN|ðχ|h⟩ in the z direction

        For the z direction χ is an m = 0 function.  The matrix element is

          √[(6/4π) ((2l+1)/(2j+1))] ⟨l,-s,1,-1|j,-1-s⟩ ⟨l,m,1,0|j,m⟩

        """
        for ell in range(ell_min, ell_max + 1):
            ellp_min = max(ell_min, ell - 1)
            ellp_max = min(ell_max, ell + 1)
            for ellp in range(ellp_min, ellp_max + 1):
                cg2 = CG(ell, -s, 1, -1, ellp, -1 - s)
                prefac = np.sqrt((2.0 * ell + 1.0) / (2.0 * ellp + 1.0))
                for m in range(-ell, ell + 1):
                    if (m < -ellp) or (m > ellp):
                        continue
                    cg1 = np.sqrt(2) * CG(ell, m, 1, 0, ellp, m)
                    yield ellp, m, ell, m, (prefac * cg1 * cg2)

    @swsh_indices_to_matrix_indices
    def ethbar_chi_z(ell_min, ell_max, s=-2):
        """Generator for the matrix element ⟨h|ð̄χ|ðN⟩ in the z direction

        For the z direction χ is an m = 0 function.  The matrix element is

          √[(6/4π) ((2l+1)/(2j+1))] ⟨l,-s-1,1,1|j,-s⟩ ⟨l,m,1,0|j,m⟩

        """
        for ell in range(ell_min, ell_max + 1):
            ellp_min = max(ell_min, ell - 1)
            ellp_max = min(ell_max, ell + 1)
            for ellp in range(ellp_min, ellp_max + 1):
                cg2 = CG(ell, -s - 1, 1, 1, ellp, -s)
                prefac = np.sqrt((2.0 * ell + 1.0) / (2.0 * ellp + 1.0))
                for m in range(-ell, ell + 1):
                    if (m < -ellp) or (m > ellp):
                        continue
                    cg1 = np.sqrt(2) * CG(ell, m, 1, 0, ellp, m)
                    yield ellp, m, ell, m, (prefac * cg1 * cg2)

    # Flux components in the x and y direction can be obtained in the plusminus
    # basis.  plus and minus corresponds to the value taken by m as +1 and -1
    # respectively.  Change of basis will give the x and y component.

    @swsh_indices_to_matrix_indices
    def eth_chi_plusminus(ell_min, ell_max, sign, s=-2):
        """Compute the ⟨ðN|ðχ|h⟩ matrix element based on the sign

        Plus and minus sign corresponds to the value taken by m=±1.

        χ is an l=1 function, and the matrix element is

          √[2(2l₁+1)(2l₂+1)/(4π(2l₃+1))] ⟨l₁,-1,l₂,-s|l₃,-1-s⟩ ⟨l₁,m₁,l₂,m₂|l₃,m₃⟩

        We use `mat_el_eth_chi` to compute the matrix elements.  Note that since
        the operator has definite m=±1, we only have mp==m±1 nonvanishing in the
        matrix elements.

        """
        if (sign != 1) and (sign != -1):
            raise ValueError("sign must be either 1 or -1 in eth_N_eth_chi_h_plusminus")

        prefac = -1.0 * sign * np.sqrt(8.0 * np.pi / 3.0)

        def mat_el_eth_chi(s, l3, m3, l1, m1, l2, m2):
            cg1 = np.sqrt(2) * CG(l1, m1, l2, m2, l3, m3)
            cg2 = CG(l1, -1.0, l2, -s, l3, -1 - s)
            return np.sqrt((2.0 * l1 + 1.0) * (2.0 * l2 + 1.0) / (4.0 * np.pi * (2.0 * l3 + 1))) * cg1 * cg2

        for ell in range(ell_min, ell_max + 1):
            ellp_min = max(ell_min, ell - 1)
            ellp_max = min(ell_max, ell + 1)
            for ellp in range(ellp_min, ellp_max + 1):
                for m in range(-ell, ell + 1):
                    mp = round(m + 1 * sign)
                    if (mp < -ellp) or (mp > ellp):
                        continue
                    yield (ellp, mp, ell, m, (prefac * mat_el_eth_chi(s, ellp, mp, 1.0, sign, ell, m)))

    eth_chi_plus = functools.partial(eth_chi_plusminus, sign=+1)
    eth_chi_minus = functools.partial(eth_chi_plusminus, sign=-1)
    eth_chi_plus.__doc__ = eth_chi_plusminus.__doc__
    eth_chi_minus.__doc__ = eth_chi_plusminus.__doc__

    @swsh_indices_to_matrix_indices
    def ethbar_chi_plusminus(ell_min, ell_max, sign, s=-2):
        """Compute the ⟨h|ðχ|ðN⟩ matrix element depending on the sign.

        Plus and minus sign corresponds to the value taken by m=±1.  Because χ is an
        l=1 function the analytical expression is

            √[2(2l₁+1)(2l₂+1)/(4π(2l₃+1))] ⟨l₁,1,l₂,-1-s|l₃,-s⟩ ⟨l₁,m₁,l₂,m₂|l₃,m₃⟩

        We use `mat_el_ethbar_chi` to compute the matrix elements.  The operator has
        definite m=±1, so we only have mp==m±1 nonvanishing in the matrix elements.

        """
        if (sign != 1) and (sign != -1):
            raise ValueError("sign must be either 1 or -1 in h_ethbar_chi_eth_N_plusminus")

        prefac = -1.0 * sign * np.sqrt(8.0 * np.pi / 3.0)

        def mat_el_ethbar_chi(s, l3, m3, l1, m1, l2, m2):
            cg1 = np.sqrt(2) * CG(l1, m1, l2, m2, l3, m3)
            cg2 = CG(l1, 1.0, l2, -1 - s, l3, -s)
            return np.sqrt((2.0 * l1 + 1.0) * (2.0 * l2 + 1.0) / (4.0 * np.pi * (2.0 * l3 + 1))) * cg1 * cg2

        for ell in range(ell_min, ell_max + 1):
            ellp_min = max(ell_min, ell - 1)
            ellp_max = min(ell_max, ell + 1)
            for ellp in range(ellp_min, ellp_max + 1):
                for m in range(-ell, ell + 1):
                    mp = round(m + 1 * sign)
                    if (mp < -ellp) or (mp > ellp):
                        continue
                    yield (ellp, mp, ell, m, (prefac * mat_el_ethbar_chi(s, ellp, mp, 1.0, sign, ell, m)))

    ethbar_chi_plus = functools.partial(ethbar_chi_plusminus, sign=+1)
    ethbar_chi_minus = functools.partial(ethbar_chi_plusminus, sign=-1)
    ethbar_chi_plus.__doc__ = ethbar_chi_plusminus.__doc__
    ethbar_chi_minus.__doc__ = ethbar_chi_plusminus.__doc__

    if not isinstance(h, WaveformModes):
        raise ValueError(
            "Boost fluxes can only be calculated from a `WaveformModes` object; "
            + "`h` is of type `{0}`.".format(type(h))
        )
    if (hdot is not None) and (not isinstance(hdot, WaveformModes)):
        raise ValueError(
            "Boost fluxes can only be calculated from a `WaveformModes` object; "
            + "`hdot` is of type `{0}`.".format(type(hdot))
        )
    if h.dataType == htype:
        if hdot is None:
            hdot = h.copy()
            hdot.dataType = hdottype
            hdot.data = h.data_dot
        elif hdot.dataType != hdottype:
            raise ValueError(
                "Input argument `hdot` is expected to have data of type `hdot`; "
                + "this `hdot` waveform data has type `{0}`".format(h.data_type_string)
            )
    else:
        raise ValueError(
            "Input argument `h` is expected to have data of type `h`; "
            + "this `h` waveform data has type `{0}`".format(h.data_type_string)
        )

    # h and h_dot are spin raised and lowered in the expression for boost flux.
    # Hence, one needs to define new waveform like object from h and hdot that
    # contains the updated data corresponding to the raising and lowering of the spin.
    # Appropriate ladder factor needs to be multiplied to each (l,m) mode.
    # eth and ethbar functions are used here for this operation.

    # eth_h --> Spin of h raised by 1; s = -1
    eth_h = h.copy()
    eth_h.data = h.eth

    # ethbar_h --> Spin of h lowered by 1; s = -3
    ethbar_h = h.copy()
    ethbar_h.data = h.ethbar

    # eth_hdot --> Spin of ḣ raised by 1; s = -1
    eth_hdot = hdot.copy()
    eth_hdot.data = hdot.eth

    # ethbar_hdot --> Spin of ḣ lowered by 1; s = -3
    ethbar_hdot = hdot.copy()
    ethbar_hdot.data = hdot.ethbar

    boost_flux = np.zeros((hdot.n_times, 3), dtype=float)

    # Each term has been computed one by taking expectation value
    # between waveform like objects (h, hdot, eth_h, ethbar_h, eth_hdot, ethbar_hdot)
    # and the matrix elements.
    # The spin of the waveform like object is taken as input for spin value s in matrix elements p_z and p_plusminus.
    # Because both ⟨a| and |b⟩ have the same spin when the matrix element is p_z and p_plusminus.

    # plus and minus terms are computed here.
    _, ethbar_hdot_chiplus_ethbar_h = matrix_expectation_value(ethbar_hdot, functools.partial(p_plus, s=-3), ethbar_h)
    _, ethbar_hdot_chiminus_ethbar_h = matrix_expectation_value(ethbar_hdot, functools.partial(p_minus, s=-3), ethbar_h)
    _, eth_hdot_chiplus_eth_h = matrix_expectation_value(eth_hdot, functools.partial(p_plus, s=-1), eth_h)
    _, eth_hdot_chiminus_eth_h = matrix_expectation_value(eth_hdot, functools.partial(p_minus, s=-1), eth_h)
    _, hdot_chiplus_h = matrix_expectation_value(hdot, functools.partial(p_plus, s=-2), h)
    _, hdot_chiminus_h = matrix_expectation_value(hdot, functools.partial(p_minus, s=-2), h)
    _, ethbar_h_chiplus_ethbar_hdot = matrix_expectation_value(ethbar_h, functools.partial(p_plus, s=-3), ethbar_hdot)
    _, ethbar_h_chiminus_ethbar_hdot = matrix_expectation_value(ethbar_h, functools.partial(p_minus, s=-3), ethbar_hdot)
    _, eth_h_chiplus_eth_hdot = matrix_expectation_value(eth_h, functools.partial(p_plus, s=-1), eth_hdot)
    _, eth_h_chiminus_eth_hdot = matrix_expectation_value(eth_h, functools.partial(p_minus, s=-1), eth_hdot)
    _, h_chiplus_hdot = matrix_expectation_value(h, functools.partial(p_plus, s=-2), hdot)
    _, h_chiminus_hdot = matrix_expectation_value(h, functools.partial(p_minus, s=-2), hdot)
    _, eth_hdot_eth_chiplus_h = matrix_expectation_value(eth_hdot, functools.partial(eth_chi_plus, s=-2), h)
    _, eth_hdot_eth_chiminus_h = matrix_expectation_value(eth_hdot, functools.partial(eth_chi_minus, s=-2), h)
    _, h_eth_chiplus_eth_hdot = matrix_expectation_value(h, functools.partial(ethbar_chi_plus, s=-2), eth_hdot)
    _, h_eth_chiminus_eth_hdot = matrix_expectation_value(h, functools.partial(ethbar_chi_minus, s=-2), eth_hdot)
    _, hdot_chiplus_hdot = matrix_expectation_value(hdot, functools.partial(p_plus, s=-2), hdot)
    _, hdot_chiminus_hdot = matrix_expectation_value(hdot, functools.partial(p_minus, s=-2), hdot)

    # Terms in the z direction are computed here.
    _, ethbar_hdot_chiz_ethbar_h = matrix_expectation_value(ethbar_hdot, functools.partial(p_z, s=-3), ethbar_h)
    _, eth_hdot_chiz_eth_h = matrix_expectation_value(eth_hdot, functools.partial(p_z, s=-1), eth_h)
    _, hdot_chiz_h = matrix_expectation_value(hdot, functools.partial(p_z, s=-2), h)
    _, ethbar_h_chiz_ethbar_hdot = matrix_expectation_value(ethbar_h, functools.partial(p_z, s=-3), ethbar_hdot)
    _, eth_h_chiz_eth_hdot = matrix_expectation_value(eth_h, functools.partial(p_z, s=-1), eth_hdot)
    _, h_chiz_hdot = matrix_expectation_value(h, functools.partial(p_z, s=-2), hdot)
    _, eth_hdot_eth_chiz_h = matrix_expectation_value(eth_hdot, functools.partial(eth_chi_z, s=-2), h)
    _, h_ethbar_chiz_eth_hdot = matrix_expectation_value(h, functools.partial(ethbar_chi_z, s=-2), eth_hdot)
    _, hdot_chiz_hdot = matrix_expectation_value(hdot, functools.partial(p_z, s=-2), hdot)

    # These terms needs to be added to get the components in a particular direction.
    # The basis will be changed from (plus, minus, z) to (x, y, z).

    boost_flux_plus = (
        (1 / 8) * (
            ethbar_hdot_chiplus_ethbar_h
            - eth_hdot_chiplus_eth_h
            + 6 * hdot_chiplus_h
            + ethbar_h_chiplus_ethbar_hdot
            - eth_h_chiplus_eth_hdot
            + 6 * h_chiplus_hdot
        )
        - (1 / 2) * np.multiply(h.t, hdot_chiplus_hdot)
        - (1 / 4) * (
            eth_hdot_eth_chiplus_h
            - h_eth_chiplus_eth_hdot
        )
    )

    boost_flux_minus = (
        (1 / 8) * (
            ethbar_hdot_chiminus_ethbar_h
            - eth_hdot_chiminus_eth_h
            + 6 * hdot_chiminus_h
            + ethbar_h_chiminus_ethbar_hdot
            - eth_h_chiminus_eth_hdot
            + 6 * h_chiminus_hdot
        )
        - (1 / 2) * np.multiply(h.t, hdot_chiminus_hdot)
        - (1 / 4) * (
            eth_hdot_eth_chiminus_h
            - h_eth_chiminus_eth_hdot
        )
    )

    # This is the component in the x direction. x = 0.5 * ( plus + minus).real
    boost_flux[:, 0] = (0.5) * (boost_flux_plus + boost_flux_minus).real

    # This is the component in the y direction. y = 0.5 * ( plus - minus).imag
    boost_flux[:, 1] = (0.5) * (boost_flux_plus - boost_flux_minus).imag

    # Component in the z direction. Only the real part of the complex value is taken into account.
    boost_flux[:, 2] = ((1 / 8) * (ethbar_hdot_chiz_ethbar_h
            - eth_hdot_chiz_eth_h
            + 6 * hdot_chiz_h
            + ethbar_h_chiz_ethbar_hdot
            - eth_h_chiz_eth_hdot
            + 6 * h_chiz_hdot)
        - (1 / 2) * np.multiply(h.t, hdot_chiz_hdot)
        + (-1 / 4) * eth_hdot_eth_chiz_h
        + (1 / 4) * h_ethbar_chiz_eth_hdot
    ).real

    # A final factor of -1/32π should be included that is present outside the integral expression of flux.
    boost_flux /= -32 * np.pi

    return boost_flux


def poincare_fluxes(h, hdot=None):
    """Compute fluxes of energy, momentum, angular momemntum, and boost

    This function will compute the time derivative 1 or 0 times (if an optional
    argument is passed) so is more efficient than separate calls to `energy_flux`,
    `momentum_flux`, `angular_momentum_flux`, and `boost_flux`.

    Parameters
    ----------
    h : WaveformModes object
        Must have type `h`

    hdot : WaveformModes object, optional [default: None]
        The time derivative of h.  If None, computed from h.  Must
        have type `hdot`.

    Returns
    -------
    edot, pdot, jdot, boost_flux : ndarray, ndarray, ndarray, ndarray

    """
    from .waveform_modes import WaveformModes
    from . import h as htype
    from . import hdot as hdottype

    if not isinstance(h, WaveformModes):
        raise ValueError(
            f"Poincare fluxes can only be calculated from a `WaveformModes` object; `h` is of type `{type(h)}`."
        )
    if (hdot is not None) and (not isinstance(hdot, WaveformModes)):
        raise ValueError(
            f"Poincare fluxes can only be calculated from a `WaveformModes` object; `hdot` is of type `{type(hdot)}`."
        )
    if h.dataType == htype:
        if hdot is None:
            hdot = h.copy()
            hdot.dataType = hdottype
            hdot.data = h.data_dot
        elif hdot.dataType != hdottype:
            raise ValueError(
                f"Input argument `hdot` is expected to have data of type `hdot`; this `hdot` waveform data has type `{h.data_type_string}`"
            )
    else:
        raise ValueError(
            f"Input argument `h` is expected to have data of type `h`; this `h` waveform data has type `{h.data_type_string}`"
        )

    return (energy_flux(hdot), momentum_flux(hdot), angular_momentum_flux(h, hdot), boost_flux(h, hdot))




import math
import numpy as np
import quaternion
import spinsfast
import spherical_functions as sf


def _process_transformation_kwargs(input_ell_max, **kwargs):
    original_kwargs = kwargs.copy()

    # Build the supertranslation and spacetime_translation arrays
    supertranslation = np.zeros((4,), dtype=complex)  # For now; may be resized below
    ell_max_supertranslation = 1  # For now; may be increased below
    if "supertranslation" in kwargs:
        supertranslation = np.array(kwargs.pop("supertranslation"), dtype=complex)
        if supertranslation.dtype != "complex" and supertranslation.size > 0:
            # I don't actually think this can ever happen...
            raise TypeError(
                "Input argument `supertranslation` should be a complex array with size>0.  "
                f"Got a {supertranslation.dtype} array of shape {supertranslation.shape}"
            )
        # Make sure the array has size at least 4, by padding with zeros
        if supertranslation.size <= 4:
            supertranslation = np.lib.pad(
                supertranslation, (0, 4 - supertranslation.size), "constant", constant_values=(0.0,)
            )
        # Check that the shape is a possible array of scalar modes with complete (ell,m) data
        ell_max_supertranslation = int(np.sqrt(len(supertranslation))) - 1
        if (ell_max_supertranslation + 1) ** 2 != len(supertranslation):
            raise ValueError(
                "Input supertranslation parameter must contain modes from ell=0 up to some ell_max, "
                "including\n           all relevant m modes in standard order (see `spherical_functions` "
                "documentation for details).\n           Thus, it must be an array with length given by a "
                "perfect square; its length is {len(supertranslation)}"
            )
        # Check that the resulting supertranslation will be real
        for ell in range(ell_max_supertranslation + 1):
            for m in range(ell + 1):
                i_pos = sf.LM_index(ell, m, 0)
                i_neg = sf.LM_index(ell, -m, 0)
                a = supertranslation[i_pos]
                b = supertranslation[i_neg]
                if abs(a - (-1.0) ** m * b.conjugate()) > 3e-16 + 1e-15 * abs(b):
                    raise ValueError(
                        f"\nsupertranslation[{i_pos}]={a}  # (ell,m)=({ell},{m})\n"
                        + "supertranslation[{}]={}  # (ell,m)=({},{})\n".format(i_neg, b, ell, -m)
                        + "Will result in an imaginary supertranslation."
                    )
    spacetime_translation = np.zeros((4,), dtype=float)
    spacetime_translation[0] = sf.constant_from_ell_0_mode(supertranslation[0]).real
    spacetime_translation[1:4] = -sf.vector_from_ell_1_modes(supertranslation[1:4]).real
    if "spacetime_translation" in kwargs:
        st_trans = np.array(kwargs.pop("spacetime_translation"), dtype=float)
        if st_trans.shape != (4,) or st_trans.dtype != "float":
            raise TypeError(
                "\nInput argument `spacetime_translation` should be a float array of shape (4,).\n"
                "Got a {} array of shape {}.".format(st_trans.dtype, st_trans.shape)
            )
        spacetime_translation = st_trans[:]
        supertranslation[0] = sf.constant_as_ell_0_mode(spacetime_translation[0])
        supertranslation[1:4] = sf.vector_as_ell_1_modes(-spacetime_translation[1:4])
    if "space_translation" in kwargs:
        s_trans = np.array(kwargs.pop("space_translation"), dtype=float)
        if s_trans.shape != (3,) or s_trans.dtype != "float":
            raise TypeError(
                "\nInput argument `space_translation` should be an array of floats of shape (3,).\n"
                "Got a {} array of shape {}.".format(s_trans.dtype, s_trans.shape)
            )
        spacetime_translation[1:4] = s_trans[:]
        supertranslation[1:4] = sf.vector_as_ell_1_modes(-spacetime_translation[1:4])
    if "time_translation" in kwargs:
        t_trans = kwargs.pop("time_translation")
        if not isinstance(t_trans, float):
            raise TypeError("Input argument `time_translation` should be a single float.  " f"Got {t_trans}")
        spacetime_translation[0] = t_trans
        supertranslation[0] = sf.constant_as_ell_0_mode(spacetime_translation[0])

    # Decide on the number of points to use in each direction.  A nontrivial supertranslation will
    # introduce power in higher modes, so for best accuracy, we need to account for that.  But we'll
    # make it a firm requirement to have enough points to capture the original waveform, at least
    output_ell_max = kwargs.pop("output_ell_max", input_ell_max)
    working_ell_max = kwargs.pop("working_ell_max", 2 * input_ell_max + ell_max_supertranslation)
    if working_ell_max < input_ell_max:
        raise ValueError(f"working_ell_max={working_ell_max} is too small; it must be at least ell_max={input_ell_max}")

    # Get the rotor for the frame rotation
    frame_rotation = np.quaternion(*np.array(kwargs.pop("frame_rotation", [1, 0, 0, 0]), dtype=float))
    if frame_rotation.abs() < 3e-16:
        raise ValueError(f"frame_rotation={frame_rotation} should be a single unit quaternion")
    frame_rotation = frame_rotation.normalized()

    # Get the boost velocity vector
    boost_velocity = np.array(kwargs.pop("boost_velocity", [0.0] * 3), dtype=float)
    beta = np.linalg.norm(boost_velocity)
    if boost_velocity.dtype != float or boost_velocity.shape != (3,) or beta >= 1.0:
        raise ValueError(
            f"Input boost_velocity=`{boost_velocity}` should be a 3-vector with " "magnitude strictly less than 1.0"
        )

    return frame_rotation, boost_velocity, supertranslation, working_ell_max, output_ell_max


def boosted_grid(frame_rotation, boost_velocity, n_theta, n_phi):
    beta = np.linalg.norm(boost_velocity)
    gamma = 1 / math.sqrt(1 - beta ** 2)
    rapidity = math.atanh(beta)

    # Construct the function that modifies our rotor grid to account for the boost
    if beta > 3e-14:  # Tolerance for beta; any smaller and numerical errors will have greater effect
        vhat = boost_velocity / beta

        def Bprm_j_k(thetaprm, phiprm):
            """Construct rotor taking r' to r

            I derived this result in a different way, but I've also found it described in
            Penrose-Rindler Vol. 1, around Eq. (1.3.5).  Note, however, that their discussion is for
            the past celestial sphere, so there's a sign difference.

            """
            # Note: It doesn't matter which we use -- r' or r; all we need is the direction of the
            # bivector spanned by v and r', which is the same as the direction of the bivector
            # spanned by v and r, since either will be normalized, and one cross product is zero iff
            # the other is zero.
            rprm = np.array(
                [math.cos(phiprm) * math.sin(thetaprm), math.sin(phiprm) * math.sin(thetaprm), math.cos(thetaprm)]
            )
            Thetaprm = math.acos(np.dot(vhat, rprm))
            Theta = 2 * math.atan(math.exp(-rapidity) * math.tan(Thetaprm / 2.0))
            rprm_cross_vhat = np.quaternion(0.0, *np.cross(rprm, vhat))
            if rprm_cross_vhat.abs() > 1e-200:
                return (rprm_cross_vhat.normalized() * (Thetaprm - Theta) / 2).exp()
            else:
                return quaternion.one

    else:

        def Bprm_j_k(thetaprm, phiprm):
            return quaternion.one

    # These are the angles in the transformed system at which we need to know the function values
    thetaprm_phiprm = sf.theta_phi(n_theta, n_phi)

    # Set up rotors that we can use to evaluate the SWSHs in the original frame
    R_j_k = np.empty((n_theta, n_phi), dtype=np.quaternion)
    for j in range(n_theta):
        for k in range(n_phi):
            thetaprm_j, phiprm_k = thetaprm_phiprm[j, k]
            R_j_k[j, k] = (
                Bprm_j_k(thetaprm_j, phiprm_k) * frame_rotation * quaternion.from_spherical_coords(thetaprm_j, phiprm_k)
            )

    return R_j_k


def conformal_factors(boost_velocity, distorted_grid_rotors):
    """Compute various combinations of the conformal factor

    This is primarily a utility function for use in the `transform` function, pulled out so that it
    can be tested separately.

    Parameters
    ==========
    boost_velocity: array of 3 floats
        Three-velocity of the new frame relative to the old frame
    distorted_grid_rotors: 2-d array of quaternions
        Unit quaternions giving the rotation of the (x, y, z) basis onto the basis vectors with
        respect to which the output spin-weighted fields are evaluated

    Returns
    =======
    k: spherical_functions.Grid
    ðk_over_k: spherical_functions.Grid
    one_over_k: spherical_functions.Grid
    one_over_k_cubed: spherical_functions.Grid
        These all have the same shape as `distorted_grid_rotors` except for an additional dimension
        of size 1 at the beginning, so that they can broadcast against the time dimension.

    """
    from quaternion import rotate_vectors

    β = np.linalg.norm(boost_velocity)
    γ = 1 / math.sqrt(1 - β ** 2)

    # Note that ðk / k = ð(v·r) / (1 - v·r), but evaluating ð(v·r) is slightly delicate.  As modes
    # in the undistorted frame, we have ð(v·r) ~ (v·r), but the right hand side is now an s=1 field,
    # so it has to be evaluated as such.
    v_dot_r = sf.Grid(np.dot(rotate_vectors(distorted_grid_rotors, quaternion.z.vec), boost_velocity), spin_weight=0)[
        np.newaxis, :, :
    ]
    ðv_dot_r = sf.Grid(
        sf.Modes(np.insert(sf.vector_as_ell_1_modes(boost_velocity), 0, 0.0), spin_weight=1).evaluate(
            distorted_grid_rotors
        ),
        spin_weight=1,
    )[np.newaxis, :, :]
    one_over_k = γ * (1 - v_dot_r)
    k = 1.0 / one_over_k
    ðk_over_k = ðv_dot_r / (1 - v_dot_r)
    one_over_k_cubed = one_over_k ** 3
    return k, ðk_over_k, one_over_k, one_over_k_cubed


def transform(self, **kwargs):
    """Apply BMS transformation to AsymptoticBondiData object

    It is important to note that the input transformation parameters are applied in this order:

      1. (Super)Translations
      2. Rotation (about the origin)
      3. Boost (about the origin)

    All input parameters refer to the transformation required to take the input data's inertial
    frame onto the inertial frame of the output data's inertial observers.  In what follows, the
    coordinates of and functions in the input inertial frame will be unprimed, while corresponding
    values of the output inertial frame will be primed.

    The translations (space, time, spacetime, or super) can be given in various ways, which may
    override each other.  Ultimately, however, they are essentially combined into a single function
    `α`, representing the supertranslation, which transforms the asymptotic time variable `u` as

        u'(u, θ, ϕ) = u(u, θ, ϕ) - α(θ, ϕ)

    A simple time translation by δt would correspond to

        α(θ, ϕ) = δt  # Independent of (θ, ϕ)

    A pure spatial translation δx would correspond to

        α(θ, ϕ) = -δx · n̂(θ, ϕ)

    where `·` is the usual dot product, and `n̂` is the unit vector in the given direction.


    Parameters
    ==========
    abd: AsymptoticBondiData
        The object storing the modes of the original data, which will be transformed in this
        function.  This is the only required argument to this function.
    time_translation: float, optional
        Defaults to zero.  Nonzero overrides corresponding components of `spacetime_translation` and
        `supertranslation` parameters.  Note that this is the actual change in the coordinate value,
        rather than the corresponding mode weight (which is what `supertranslation` represents).
    space_translation : float array of length 3, optional
        Defaults to empty (no translation).  Non-empty overrides corresponding components of
        `spacetime_translation` and `supertranslation` parameters.  Note that this is the actual
        change in the coordinate value, rather than the corresponding mode weight (which is what
        `supertranslation` represents).
    spacetime_translation : float array of length 4, optional
        Defaults to empty (no translation).  Non-empty overrides corresponding components of
        `supertranslation`.  Note that this is the actual change in the coordinate value, rather
        than the corresponding mode weight (which is what `supertranslation` represents).
    supertranslation : complex array [defaults to 0]
        This gives the complex components of the spherical-harmonic expansion of the
        supertranslation in standard form, starting from ell=0 up to some ell_max, which may be
        different from the ell_max of the input `abd` object.  Supertranslations must be real, so
        these values should obey the condition
            α^{ℓ,m} = (-1)^m ᾱ^{ℓ,-m}
        This condition is actually imposed on the input data, so imaginary parts of α(θ, ϕ) will
        essentially be discarded.  Defaults to empty, which causes no supertranslation.  Note that
        some components may be overridden by the parameters above.
    frame_rotation : quaternion [defaults to 1]
        Transformation applied to (x,y,z) basis of the input mode's inertial frame.  For example,
        the basis z vector of the new frame may be written as
           z' = frame_rotation * z * frame_rotation.inverse()
        Defaults to 1, corresponding to the identity transformation (no rotation).
    boost_velocity : float array of length 3 [defaults to (0, 0, 0)]
        This is the three-velocity vector of the new frame relative to the input frame.  The norm of
        this vector is required to be smaller than 1.
    output_ell_max: int [defaults to abd.ell_max]
        Maximum ell value in the output data.
    working_ell_max: int [defaults to 2 * abd.ell_max]
        Maximum ell value to use during the intermediate calculations.  Rotations and time
        translations do not require this to be any larger than abd.ell_max, but other
        transformations will require more values of ell for accurate results.  In particular, boosts
        are multiplied by time, meaning that a large boost of data with large values of time will
        lead to very large power in higher modes.  Similarly, large (super)translations will couple
        power through a lot of modes.  To avoid aliasing, this value should be large, to accomodate
        power in higher modes.

    Returns
    -------
    abdprime: AsymptoticBondiData
        Object representing the transformed data.

    """
    from quaternion import rotate_vectors
    from scipy.interpolate import CubicSpline

    # Parse the input arguments, and define the basic parameters for this function
    (
        frame_rotation,
        boost_velocity,
        supertranslation,
        working_ell_max,
        output_ell_max,
    ) = _process_transformation_kwargs(self.ell_max, **kwargs)
    n_theta = 2 * working_ell_max + 1
    n_phi = n_theta
    β = np.linalg.norm(boost_velocity)
    γ = 1 / math.sqrt(1 - β ** 2)

    # Make this into a Modes object, so it can keep track of its spin weight, etc., through the
    # various operations needed below.
    supertranslation = sf.Modes(supertranslation, spin_weight=0).real

    # This is a 2-d array of unit quaternions, which are what the spin-weighted functions should be
    # evaluated on (even for spin 0 functions, for simplicity).  That will be equivalent to
    # evaluating the spin-weighted functions with respect to the transformed grid -- although on the
    # original time slices.
    distorted_grid_rotors = boosted_grid(frame_rotation, boost_velocity, n_theta, n_phi)

    # Compute u, α, ðα, ððα, k, ðk/k, 1/k, and 1/k³ on the distorted grid, including new axes to
    # enable broadcasting with time-dependent functions.  Note that the first axis should represent
    # variation in u, the second axis variation in θ', and the third axis variation in ϕ'.
    u = self.u
    α = sf.Grid(supertranslation.evaluate(distorted_grid_rotors), spin_weight=0).real[np.newaxis, :, :]
    # The factors of 1/sqrt(2) and 1/2 come from using the GHP eth instead of the NP eth.
    ðα = sf.Grid(supertranslation.eth.evaluate(distorted_grid_rotors) / np.sqrt(2), spin_weight=α.s + 1)[np.newaxis, :, :]
    ððα = sf.Grid(0.5 * supertranslation.eth.eth.evaluate(distorted_grid_rotors), spin_weight=α.s + 2)[np.newaxis, :, :]
    k, ðk_over_k, one_over_k, one_over_k_cubed = conformal_factors(boost_velocity, distorted_grid_rotors)

    # ðu'(u, θ', ϕ') exp(iλ) / k(θ', ϕ')
    ðuprime_over_k = ðk_over_k * (u - α) - ðα

    # ψ0(u, θ', ϕ') exp(2iλ)
    ψ0 = sf.Grid(self.psi0.evaluate(distorted_grid_rotors), spin_weight=2)
    # ψ1(u, θ', ϕ') exp(iλ)
    ψ1 = sf.Grid(self.psi1.evaluate(distorted_grid_rotors), spin_weight=1)
    # ψ2(u, θ', ϕ')
    ψ2 = sf.Grid(self.psi2.evaluate(distorted_grid_rotors), spin_weight=0)
    # ψ3(u, θ', ϕ') exp(-1iλ)
    ψ3 = sf.Grid(self.psi3.evaluate(distorted_grid_rotors), spin_weight=-1)
    # ψ4(u, θ', ϕ') exp(-2iλ)
    ψ4 = sf.Grid(self.psi4.evaluate(distorted_grid_rotors), spin_weight=-2)
    # σ(u, θ', ϕ') exp(2iλ)
    σ = sf.Grid(self.sigma.evaluate(distorted_grid_rotors), spin_weight=2)

    ### The following calculations are done using in-place Horner form.  I suspect this will be the
    ### most efficient form of this calculation, within reason.  Note that the factors of exp(isλ)
    ### were computed automatically by evaluating in terms of quaternions.
    #
    fprime_of_timenaught_directionprime = np.empty((6, self.n_times, n_theta, n_phi), dtype=complex)
    # ψ0'(u, θ', ϕ')
    fprime_temp = ψ4.copy()
    fprime_temp *= ðuprime_over_k
    fprime_temp += -4 * ψ3
    fprime_temp *= ðuprime_over_k
    fprime_temp += 6 * ψ2
    fprime_temp *= ðuprime_over_k
    fprime_temp += -4 * ψ1
    fprime_temp *= ðuprime_over_k
    fprime_temp += ψ0
    fprime_temp *= one_over_k_cubed
    fprime_of_timenaught_directionprime[0] = fprime_temp
    # ψ1'(u, θ', ϕ')
    fprime_temp = -ψ4
    fprime_temp *= ðuprime_over_k
    fprime_temp += 3 * ψ3
    fprime_temp *= ðuprime_over_k
    fprime_temp += -3 * ψ2
    fprime_temp *= ðuprime_over_k
    fprime_temp += ψ1
    fprime_temp *= one_over_k_cubed
    fprime_of_timenaught_directionprime[1] = fprime_temp
    # ψ2'(u, θ', ϕ')
    fprime_temp = ψ4.copy()
    fprime_temp *= ðuprime_over_k
    fprime_temp += -2 * ψ3
    fprime_temp *= ðuprime_over_k
    fprime_temp += ψ2
    fprime_temp *= one_over_k_cubed
    fprime_of_timenaught_directionprime[2] = fprime_temp
    # ψ3'(u, θ', ϕ')
    fprime_temp = -ψ4
    fprime_temp *= ðuprime_over_k
    fprime_temp += ψ3
    fprime_temp *= one_over_k_cubed
    fprime_of_timenaught_directionprime[3] = fprime_temp
    # ψ4'(u, θ', ϕ')
    fprime_temp = ψ4.copy()
    fprime_temp *= one_over_k_cubed
    fprime_of_timenaught_directionprime[4] = fprime_temp
    # σ'(u, θ', ϕ')
    fprime_temp = σ.copy()
    fprime_temp -= ððα
    fprime_temp *= one_over_k
    fprime_of_timenaught_directionprime[5] = fprime_temp

    # Determine the new time slices.  The set timeprime is chosen so that on each slice of constant
    # u'_i, the average value of u=(u'/k)+α is precisely <u>=u'γ+<α>=u_i.  But then, we have to
    # narrow that set down, so that every grid point on all the u'_i' slices correspond to data in
    # the range of input data.
    timeprime = (u - sf.constant_from_ell_0_mode(supertranslation[0]).real) / γ
    timeprime_of_initialtime_directionprime = k * (u[0] - α)
    timeprime_of_finaltime_directionprime = k * (u[-1] - α)
    earliest_complete_timeprime = np.max(timeprime_of_initialtime_directionprime.view(np.ndarray))
    latest_complete_timeprime = np.min(timeprime_of_finaltime_directionprime.view(np.ndarray))
    timeprime = timeprime[(timeprime >= earliest_complete_timeprime) & (timeprime <= latest_complete_timeprime)]

    # This will store the values of f'(u', θ', ϕ') for the various functions `f`
    fprime_of_timeprime_directionprime = np.zeros((6, timeprime.size, n_theta, n_phi), dtype=complex)

    # Interpolate the various transformed function values on the transformed grid from the original
    # time coordinate to the new set of time coordinates, independently for each direction.
    for i in range(n_theta):
        for j in range(n_phi):
            k_i_j = k[0, i, j]
            α_i_j = α[0, i, j]
            # u'(u, θ', ϕ')
            timeprime_of_timenaught_directionprime_i_j = k_i_j * (u - α_i_j)
            # f'(u', θ', ϕ')
            fprime_of_timeprime_directionprime[:, :, i, j] = CubicSpline(
                timeprime_of_timenaught_directionprime_i_j, fprime_of_timenaught_directionprime[:, :, i, j], axis=1
            )(timeprime)

    # Finally, transform back from the distorted grid to the SWSH mode weights as measured in that
    # grid.  I'll abuse notation slightly here by indicating those "distorted" mode weights with
    # primes, so that f'(u')_{ℓ', m'} = ∫ f'(u', θ', ϕ') sȲ_{ℓ', m'}(θ', ϕ') sin(θ') dθ' dϕ'
    abdprime = type(self)(timeprime, output_ell_max)
    # ψ0'(u')_{ℓ', m'}
    abdprime.psi0 = spinsfast.map2salm(fprime_of_timeprime_directionprime[0], 2, output_ell_max)
    # ψ1'(u')_{ℓ', m'}
    abdprime.psi1 = spinsfast.map2salm(fprime_of_timeprime_directionprime[1], 1, output_ell_max)
    # ψ2'(u')_{ℓ', m'}
    abdprime.psi2 = spinsfast.map2salm(fprime_of_timeprime_directionprime[2], 0, output_ell_max)
    # ψ3'(u')_{ℓ', m'}
    abdprime.psi3 = spinsfast.map2salm(fprime_of_timeprime_directionprime[3], -1, output_ell_max)
    # ψ4'(u')_{ℓ', m'}
    abdprime.psi4 = spinsfast.map2salm(fprime_of_timeprime_directionprime[4], -2, output_ell_max)
    # σ'(u')_{ℓ', m'}
    abdprime.sigma = spinsfast.map2salm(fprime_of_timeprime_directionprime[5], 2, output_ell_max)

    return abdprime



# Copyright (c) 2021, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>
"""Submodule for operating on SpEC waveform files"""

from .. import __version__

from .com_motion import com_motion, estimate_avg_com_motion, remove_avg_com_motion
from .file_io import read_from_h5, write_to_h5, create_abd_from_h5, rpxmb



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

import warnings
import os
import numpy as np
import quaternion
import spherical_functions as sf
from ... import WaveformModes, DataType, DataNames

sxs_formats = [
    "corotating_paired_xor",
]


def save(w, file_name=None, L2norm_fractional_tolerance=1e-10, log_frame=None, compress=True):
    import tempfile
    import contextlib
    import pathlib
    import json
    import numpy
    import scipy
    import h5py
    import sxs
    import scri
    from ...utilities import xor_timeseries, fletcher32

    # Make sure that we can understand the file_name and create the directory
    if file_name is None:
        # We'll just be creating a temp directory below, to check
        warnings.warn(
            "Input `file_name` is None.  Running in temporary directory.\n"
            "Note that this option is mostly for debugging purposes."
        )
    else:
        h5_path = pathlib.Path(file_name).expanduser().resolve().with_suffix(".h5")
        if not h5_path.parent.exists():
            h5_path.parent.mkdir(parents=True)

    compression_options = (
        {"compression": "gzip", "compression_opts": 9, "shuffle": True, "fletcher32": True,} if compress else {}
    )

    if L2norm_fractional_tolerance == 0.0:
        log_frame = quaternion.as_float_array(np.log(w.frame))[:, 1:]
    else:
        # We need this storage anyway, so let's just make a copy and work in-place
        w = w.copy()
        if log_frame is not None:
            log_frame = log_frame.copy()

        # Ensure waveform is in corotating frame
        if w.frameType == scri.Inertial:
            try:
                initial_time = w.t[0]
                relaxation_time = w.metadata.relaxation_time
                max_norm_time = w.max_norm_time()
                z_alignment_region = ((relaxation_time - initial_time) / (max_norm_time - initial_time), 0.95)
            except:
                z_alignment_region = (0.1, 0.95)
            w, log_frame = w.to_corotating_frame(
                tolerance=1e-10, z_alignment_region=z_alignment_region, truncate_log_frame=True
            )
            log_frame = log_frame[:, 1:]
        if w.frameType != scri.Corotating:
            raise ValueError(
                "Frame type of input waveform must be 'Corotating' or 'Inertial'; " f"it is {w.frame_type_string}"
            )

        # Convert mode structure to conjugate pairs
        w.convert_to_conjugate_pairs()

        # Set bits below the desired significance level to 0
        w.truncate(tol=L2norm_fractional_tolerance)

        # Compute log(frame)
        if log_frame is None:
            log_frame = quaternion.as_float_array(np.log(w.frame))[:, 1:]
            power_of_2 = 2 ** (-np.floor(np.log2(L2norm_fractional_tolerance / 10))).astype("int")
            log_frame = np.round(log_frame * power_of_2) / power_of_2

        # Change -0.0 to 0.0 (~.5% compression for non-precessing systems)
        w.t += 0.0
        w.data += 0.0
        log_frame += 0.0

        # XOR successive instants in time
        xor_timeseries(w.t)
        xor_timeseries(w.data)
        xor_timeseries(log_frame)

    # Make sure we have a place to keep all this
    with contextlib.ExitStack() as context:
        if file_name is None:
            temp_dir = context.enter_context(tempfile.TemporaryDirectory())
            h5_path = pathlib.Path(f"{temp_dir}") / "test.h5"
        else:
            print(f'Saving H5 to "{h5_path}"')

        # Write the H5 file
        with h5py.File(h5_path, "w") as f:
            f.attrs["sxs_format"] = "corotating_paired_xor"
            warnings.warn('sxs_format is being set to "corotating_paired_xor"')
            f.create_dataset("time", data=w.t.view(np.uint64), chunks=(w.n_times,), **compression_options)
            f.create_dataset("modes", data=w.data.view(np.uint64), chunks=(w.n_times, 1), **compression_options)
            f["modes"].attrs["ell_min"] = w.ell_min
            f["modes"].attrs["ell_max"] = w.ell_max
            if log_frame.size > 1:
                f.create_dataset(
                    "log_frame", data=log_frame.view(np.uint64), chunks=(w.n_times, 1), **compression_options
                )

        # Get some numbers for the JSON file
        h5_size = os.stat(h5_path).st_size
        if file_name is None:
            print(f"Output H5 file size: {h5_size:_} B")
        fletcher32_dict = {}
        with h5py.File(h5_path, "r") as f:
            fletcher32_dict["time"] = fletcher32(f["time"][:])
            fletcher32_dict["modes"] = fletcher32(f["modes"][:])
            if "log_frame" in f:
                fletcher32_dict["log_frame"] = fletcher32(f["log_frame"][:])

        # Write the corresponding JSON file
        json_path = h5_path.with_suffix(".json")
        json_data = {
            "sxs_format": "corotating_paired_xor",
            "data_info": {
                "data_type": w.data_type_string,
                "spin_weight": int(w.spin_weight),
                "ell_min": int(w.ell_min),
                "ell_max": int(w.ell_max),
            },
            "transformations": {
                "truncation": L2norm_fractional_tolerance,
                # see below for 'boost_velocity'
                # see below for 'space_translation'
            },
            "version_info": {
                "numpy": numpy.__version__,
                "scipy": scipy.__version__,
                "h5py": h5py.__version__,
                "quaternion": quaternion.__version__,
                "spherical_functions": sf.__version__,
                "scri": scri.__version__,
                "sxs": sxs.__version__,
                # see below 'spec_version_hist'
            },
            "validation": {"h5_file_size": h5_size, "n_times": w.n_times, "fletcher32": fletcher32_dict,},
        }
        if hasattr(w, "boost_velocity"):
            json_data["transformations"]["boost_velocity"] = w.boost_velocity.tolist()
        if hasattr(w, "space_translation"):
            json_data["transformations"]["space_translation"] = w.space_translation.tolist()
        if hasattr(w, "version_hist"):
            json_data["version_info"]["spec_version_history"] = w.version_hist
        if file_name is not None:
            print(f'Saving JSON to "{json_path}"')
        with json_path.open("w") as f:
            json.dump(json_data, f, indent=2, separators=(",", ": "), ensure_ascii=True)

    return w


def load(file_name, ignore_validation=False):
    import pathlib
    import json
    import h5py
    import scri
    from ...utilities import xor_timeseries_reverse, fletcher32

    def invalid(message):
        if ignore_validation:
            warnings.warn(message)
        else:
            raise ValueError(message)

    h5_path = pathlib.Path(file_name).expanduser().resolve().with_suffix(".h5")
    json_path = h5_path.with_suffix(".json")

    # This will be used for validation
    h5_size = os.stat(h5_path).st_size

    if not json_path.exists():
        invalid(f'JSON file "{json_path}" cannot be found, but is expected for this data format.')
        json_data = {}
    else:
        with open(json_path) as f:
            json_data = json.load(f)

        dataType = json_data.get("data_info", {}).get("data_type", "UnknownDataType")
        dataType = scri.DataType[scri.DataNames.index(dataType)]

        # Make sure this is our format
        sxs_format = json_data.get("sxs_format", "")
        if sxs_format not in sxs_formats:
            invalid(
                f'The `sxs_format` found in JSON file is "{sxs_format}"; it should be one of\n' f"    {sxs_formats}."
            )

        json_h5_file_size = json_data.get("validation", {}).get("h5_file_size", 0)
        if json_h5_file_size != h5_size:
            invalid(
                f"Mismatch between `validation/h5_file_size` key in JSON file ({json_h5_file_size}) "
                f'and observed file size ({h5_size}) of "{h5_path}".'
            )

    with h5py.File(h5_path, "r") as f:
        # Make sure this is our format
        sxs_format = f.attrs["sxs_format"]
        if sxs_format not in sxs_formats:
            raise ValueError(
                f'The `sxs_format` found in H5 file is "{sxs_format}"; it should be one of\n' f"    {sxs_formats}."
            )

        # Ensure that the 'validation' keys from the JSON file are the same as in this file
        json_n_times = json_data.get("validation", {}).get("n_times", 0)
        n_times = f["time"][:].view(np.float64).size
        if json_n_times != n_times:
            invalid(
                f"Number of time steps in H5 file ({n_times}) "
                f"does not match expected value from JSON ({json_n_times})."
            )
        for dataset, checksum in json_data.get("validation", {}).get("fletcher32", {}).items():
            observed_checksum = fletcher32(f[dataset][:])
            if checksum != observed_checksum:
                invalid(f'Checksum of "{dataset}" dataset does not match expected value from JSON.')

        # Read the data
        time = f["time"][:].view(np.float64)
        modes = f["modes"][:].view(complex)
        ell_min = f["modes"].attrs["ell_min"]
        ell_max = f["modes"].attrs["ell_max"]
        if "log_frame" in f:
            log_frame = f["log_frame"][:].view(np.float64)
        else:
            log_frame = np.empty((0, 3), dtype=np.float64)

    xor_timeseries_reverse(time)
    xor_timeseries_reverse(modes)
    xor_timeseries_reverse(log_frame)
    frame = np.exp(quaternion.as_quat_array(np.insert(log_frame, 0, 0.0, axis=1)))

    w = WaveformModes(
        t=time,
        frame=frame,
        data=modes,
        frameType=scri.Corotating,
        dataType=dataType,
        m_is_scaled_out=True,
        r_is_scaled_out=True,
        ell_min=ell_min,
        ell_max=ell_max,
    )
    w.convert_from_conjugate_pairs()
    w.json_data = json_data

    return w



import sys
from os import walk
from os.path import join, exists
from fnmatch import filter
from multiprocessing import Pool
from scri.SpEC import remove_avg_com_motion as racm


def run_in(args):
    (filename, i_this, i_tot) = args
    if exists(filename.replace(".h5", "_CoM.h5")):
        return
    print(f"Working on {filename} -- {i_this} of {i_tot}")
    sys.stdout.flush()
    try:
        racm(f"{filename}/Extrapolated_N2.dir", plot=True, file_write_mode="w")
        try:
            racm(f"{filename}/Extrapolated_N3.dir", plot=True, file_write_mode="a")
        except:
            print(f"Failed in {filename}/Extrapolated_N3.dir -- {i_this} of {i_tot}")
            sys.stdout.flush()
        try:
            racm(f"{filename}/Extrapolated_N4.dir", plot=True, file_write_mode="a")
        except:
            print(f"Failed in {filename}/Extrapolated_N4.dir -- {i_this} of {i_tot}")
            sys.stdout.flush()
        try:
            racm(f"{filename}/OutermostExtraction.dir", plot=True, file_write_mode="a")
        except:
            print(f"Failed in {filename}/OutermostExtraction.dir -- {i_this} of {i_tot}")
            sys.stdout.flush()
    except:
        print(f"Failed in {filename} -- {i_this} of {i_tot}")
        sys.stdout.flush()
    print(f"Finished {filename} -- {i_this} of {i_tot}")
    sys.stdout.flush()


if __name__ == "__main__":
    print("Finding files to operate on")
    files = [
        join(root, "rhOverM_Asymptotic_GeometricUnits.h5")
        for top_dir in ["Catalog", "Incoming"]
        for root, dirnames, filenames in walk(top_dir)
        for filename in filter(filenames, "rhOverM_Asymptotic_GeometricUnits.h5")
    ]
    print("Finished finding files to operate on")
    pool = Pool(processes=12)
    pool.map(run_in, zip(files, range(1, len(files) + 1), [len(files)] * len(files)))



# Copyright (c) 2021, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

sxs_formats = [
    "rotating_paired_xor_multishuffle_bzip2",
]
default_shuffle_widths = (8, 8, 4, 4, 4, 2,) + (1,) * 34


def save(w, file_name=None, file_write_mode="w", L2norm_fractional_tolerance=1e-10, log_frame=None, shuffle_widths=default_shuffle_widths):
    """Save a waveform in RPXMB format

    This function converts the data to "rotating paired XOR multishuffle bzip2"
    format.  In particular, it uses the corotating frame, and zeroes out bits at
    high precision to allow for optimal compression while maintaining the requested
    tolerance.

    Parameters
    ----------
    w : WaveformModes
        A waveform in either the inertial or corotating frame
    file_name : str
        Relative or absolute path to the output HDF5 file.  If this string contains
        `'.h5'` but does not *end* with that, the remainder of the string is taken
        to be the group within the HDF5 file in which the data should be stored.
        Also note that a JSON file is created in the same location, with `.h5`
        replaced by `.json` (and the corresponding data is stored under the `group`
        key if relevant).  For testing purposes, this argument may be `None`, in
        which case a temporary directory is used, just to test how large the output
        will be; it is deleted immediately upon returning.
    file_write_mode : str, optional
        One of the valid [file modes for
        h5py](https://docs.h5py.org/en/stable/high/file.html#opening-creating-files).
        Default value is `"w"`, which overwrites any existing file.  If writing
        into a group, you may prefer `"a"`, which will just ensure the file exists
        without erasing it first.
    L2norm_fractional_tolerance : float, optional
        Tolerance passed to `WaveformModes.truncate`; see that function's docstring
        for details.  Default value is 1e-10.
    log_frame : array of quaternions, optional
        If this argument is given the waveform must be in the corotating frame, and
        the given data will be used as the logarithmic frame data.  If this
        argument is `None` (the default), this will be calculated when the waveform
        is transformed to the corotating frame, or simply taken directly from the
        waveform if it is already corotating.
    shuffle_widths : iterable of ints, optional
        See `scri.utilities.multishuffle` for details.  The default value is
        `default_shuffle_widths`.  Note that if `L2norm_fractional_tolerance` is
        0.0, this will be ignored and the standard HDF5 shuffle option will be used
        instead.

    Returns
    -------
    w_out : WaveformModes
        The output data, after conversion to the corotating frame, pairing of
        opposite `m` modes, and XOR-ing (but not shuffling).
    log_frame : array of quaternions
        The actual `log_frame` data stored in the file, and used to transform to
        the corotating frame if that was done inside this function.

    """
    import sys
    import os
    import warnings
    import tempfile
    import contextlib
    import pathlib
    import bz2
    import json
    import numpy as np
    import scipy
    import h5py
    import quaternion
    import sxs
    import scri
    import spherical_functions as sf
    from scri.utilities import xor_timeseries as xor
    from sxs.utilities import md5checksum

    # Make sure that we can understand the file_name and create the directory
    group = None
    if file_name is None:
        # We'll just be creating a temp directory below, to check
        warnings.warn(
            "\nInput `file_name` is None.  Running in temporary directory.\n"
            "Note that this option is mostly for debugging purposes."
        )
    else:
        if ".h5" in file_name and not file_name.endswith(".h5"):
            file_name, group = file_name.split(".h5")
        h5_path = pathlib.Path(file_name).expanduser().resolve().with_suffix(".h5")
        if not h5_path.parent.exists():
            h5_path.parent.mkdir(parents=True)
    if group == "/":
        group = None

    shuffle = scri.utilities.multishuffle(tuple(shuffle_widths))

    if L2norm_fractional_tolerance == 0.0:
        log_frame = quaternion.as_float_array(np.log(w.frame))[:, 1:]
    else:
        # We need this storage anyway, so let's just make a copy and work in-place
        w = w.copy()
        if log_frame is not None:
            log_frame = log_frame.copy()

        # Ensure waveform is in corotating frame
        if w.frameType == scri.Inertial:
            try:
                initial_time = w.t[0]
                relaxation_time = w.metadata.relaxation_time
                max_norm_time = w.max_norm_time()
                z_alignment_region = ((relaxation_time - initial_time) / (max_norm_time - initial_time), 0.95)
            except:
                z_alignment_region = (0.1, 0.95)
            w, log_frame = w.to_corotating_frame(
                tolerance=1e-10, z_alignment_region=z_alignment_region, truncate_log_frame=True
            )
            log_frame = log_frame[:, 1:]
        if w.frameType != scri.Corotating:
            raise ValueError(
                "Frame type of input waveform must be 'Corotating' or 'Inertial'; " f"it is {w.frame_type_string}"
            )

        # Convert mode structure to conjugate pairs
        w.convert_to_conjugate_pairs()

        # Set bits below the desired significance level to 0
        w.truncate(tol=L2norm_fractional_tolerance)

        # Compute log(frame)
        if log_frame is None:
            log_frame = quaternion.as_float_array(np.log(w.frame))[:, 1:]
            # power_of_2 = 2 ** (-np.floor(np.log2(L2norm_fractional_tolerance / 10))).astype("int")
            # log_frame = np.round(log_frame * power_of_2) / power_of_2

        # Change -0.0 to 0.0 (~.5% compression for non-precessing systems)
        w.t += 0.0
        w.data += 0.0
        log_frame += 0.0

        # XOR successive instants in time
        xor(w.t)
        xor(w.data)
        xor(log_frame)

    # Make sure we have a place to keep all this
    with contextlib.ExitStack() as context:
        if file_name is None:
            temp_dir = context.enter_context(tempfile.TemporaryDirectory())
            h5_path = pathlib.Path(f"{temp_dir}") / "test.h5"
        else:
            print(f'Saving H5 to "{h5_path}"')

        # Write the H5 file
        with h5py.File(h5_path, file_write_mode) as f:
            # If we are writing to a group within the file, create it
            if group is not None:
                g = f.create_group(group)
            else:
                g = f
            if L2norm_fractional_tolerance != 0.0:
                g.attrs["sxs_format"] = f"{sxs_formats[0]}"
                g.attrs["n_times"] = w.n_times
                g.attrs["ell_min"] = w.ell_min
                g.attrs["ell_max"] = w.ell_max
                g.attrs["shuffle_widths"] = np.array(shuffle_widths, dtype=np.uint8)
                # warnings.warn(f'sxs_format is being set to "{sxs_formats[0]}"')
                data = np.void(
                    bz2.compress(
                        shuffle(w.t.view(np.uint64)).tobytes()
                        + shuffle(w.data.view(np.uint64).flatten("F")).tobytes()
                        + shuffle(log_frame.view(np.uint64).flatten("F")).tobytes()
                    )
                )
                g.create_dataset("data", data=data)
            else:
                compression_options = {
                    "compression": "gzip",
                    "compression_opts": 9,
                    "shuffle": True,
                }
                g.attrs["sxs_format"] = f"{sxs_formats[0]}"
                g.create_dataset("time", data=w.t.view(np.uint64), chunks=(w.n_times,), **compression_options)
                g.create_dataset("modes", data=w.data.view(np.uint64), chunks=(w.n_times, 1), **compression_options)
                g["modes"].attrs["ell_min"] = w.ell_min
                g["modes"].attrs["ell_max"] = w.ell_max
                if log_frame.size > 1:
                    g.create_dataset(
                        "log_frame", data=log_frame.view(np.uint64), chunks=(w.n_times, 1), **compression_options
                    )

        # Get some numbers for the JSON file
        h5_size = os.stat(h5_path).st_size
        if file_name is None:
            print(f"Output H5 file size: {h5_size:_} B")
        md5sum = md5checksum(h5_path)

        if file_name is not None:
            # Set up the corresponding JSON information
            json_data = {
                "sxs_format": sxs_formats[0],
                "data_info": {
                    "data_type": w.data_type_string,
                    "m_is_scaled_out": w.m_is_scaled_out,
                    "r_is_scaled_out": w.r_is_scaled_out,
                    "spin_weight": int(w.spin_weight),
                    "ell_min": int(w.ell_min),
                    "ell_max": int(w.ell_max),
                },
                "transformations": {
                    "truncation": L2norm_fractional_tolerance,
                    # see below for 'boost_velocity'
                    # see below for 'space_translation'
                },
                "version_info": {
                    "python": sys.version,
                    "numpy": np.__version__,
                    "scipy": scipy.__version__,
                    "h5py": h5py.__version__,
                    "quaternion": quaternion.__version__,
                    "spherical_functions": sf.__version__,
                    "scri": scri.__version__,
                    "sxs": sxs.__version__,
                    # see below 'spec_version_hist'
                },
                # see below for 'validation'
            }
            if group is not None:
                json_data["validation"] = {
                    "n_times": w.n_times,
                }
            else:
                json_data["validation"] = {
                    "h5_file_size": h5_size,
                    "n_times": w.n_times,
                    "md5sum": md5sum
                }
            if hasattr(w, "boost_velocity"):
                json_data["transformations"]["boost_velocity"] = w.boost_velocity.tolist()
            if hasattr(w, "space_translation"):
                json_data["transformations"]["space_translation"] = w.space_translation.tolist()
            if hasattr(w, "version_hist"):
                json_data["version_info"]["spec_version_history"] = w.version_hist

            # Write the corresponding JSON file
            json_path = h5_path.with_suffix(".json")
            print(f'Saving JSON to "{json_path}"')
            if group is not None:
                if json_path.exists() and file_write_mode!="w":
                    with json_path.open("r") as f:
                        original_json = json.load(f)
                else:
                    original_json = {}
                original_json[group] = json_data
                json_data = original_json
            with json_path.open("w") as f:
                json.dump(json_data, f, indent=2, separators=(",", ": "), ensure_ascii=True)

    return w, log_frame


def load(file_name, ignore_validation=False, check_md5=True, **kwargs):
    """Load a waveform in RPXMB format

    Parameters
    ----------
    file_name : str
        Relative or absolute path to the input HDF5 file.  If this string contains
        but does not *end* with `'.h5'`, the remainder of the string is taken to be
        the group within the HDF5 file in which the data is stored.  Also note that
        a JSON file is expected in the same location, with `.h5` replaced by
        `.json` (and the corresponding data must be stored under the `group` key if
        relevant).
    ignore_validation : bool, optional
        If `True`, the JSON file need not be present, and the validation keys
        (`h5_file_size`, `n_times`, and `md5sum`) will be ignored — though warnings
        may be issued.  If `False`, these are all required, with the possible
        exception of `h5_file_size` and `md5sum` if a group is used within the HDF5
        file, or `md5sum` if `check_md5` is `False`.
    check_md5 : bool, optional
        Default is `True`.  See `ignore_validation` for explanation.

    Keyword parameters
    ------------------
    data_type : str, optional
        One of `scri.DataNames`.  Default is "UnknownDataType".
    m_is_scaled_out : bool, optional
        Default is True
    r_is_scaled_out : bool, optional
        Default is True

    Note that the keyword parameters will be overridden by corresponding entries in
    the JSON file, if they exist.  If the JSON file does not exist, any keyword
    parameters not listed above will be passed through as the `json_data` field of
    the returned waveform.

    """
    import os
    import warnings
    import pathlib
    import bz2
    import json
    import numpy as np
    import h5py
    import quaternion
    import scri
    from scri.utilities import xor_timeseries_reverse as unxor
    from sxs.utilities import md5checksum

    def invalid(message):
        if ignore_validation:
            pass
        elif ignore_validation is None:
            warnings.warn(message)
        else:
            raise ValueError(message)

    group = None
    if ".h5" in file_name and not file_name.endswith(".h5"):
        file_name, group = file_name.split(".h5")
    if group == "/":
        group = None

    h5_path = pathlib.Path(file_name).expanduser().resolve().with_suffix(".h5")
    json_path = h5_path.with_suffix(".json")

    # This will be used for validation
    h5_size = h5_path.stat().st_size

    data_type = kwargs.pop("data_type", "UnknownDataType")
    m_is_scaled_out = bool(kwargs.pop("m_is_scaled_out", True))
    r_is_scaled_out = bool(kwargs.pop("r_is_scaled_out", True))

    if not json_path.exists():
        invalid(f'\nJSON file "{json_path}" cannot be found, but is expected for this data format.')
        json_data = kwargs.copy()
    else:
        with open(json_path) as f:
            json_data = json.load(f)
        if group is not None:
            json_data = json_data[group]

        data_type = json_data.get("data_info", {}).get("data_type", data_type)
        m_is_scaled_out = bool(json_data.get("data_info", {}).get("m_is_scaled_out", m_is_scaled_out))
        r_is_scaled_out = bool(json_data.get("data_info", {}).get("r_is_scaled_out", r_is_scaled_out))

        # Make sure this is our format
        sxs_format = json_data.get("sxs_format", "")
        if sxs_format not in sxs_formats:
            invalid(
                f"\nThe `sxs_format` found in JSON file is '{sxs_format}';\n"
                f"it should be one of\n"
                f"    {sxs_formats}."
            )

        if group is None:
            # Make sure the expected H5 file size matches the observed value
            json_h5_file_size = json_data.get("validation", {}).get("h5_file_size", 0)
            if json_h5_file_size != h5_size:
                invalid(
                    f"\nMismatch between `validation/h5_file_size` key in JSON file ({json_h5_file_size}) "
                    f'and observed file size ({h5_size}) of "{h5_path}".'
                )

            # Make sure the expected H5 file hash matches the observed value
            if check_md5:
                md5sum = md5checksum(h5_path)
                json_md5sum = json_data.get("validation", {}).get("md5sum", "")
                if json_md5sum != md5sum:
                    invalid(f"\nMismatch between `validation/md5sum` key in JSON file and observed MD5 checksum.")

    dataType = scri.DataType[scri.DataNames.index(data_type)]

    with h5py.File(h5_path, "r") as f:
        if group is not None:
            g = f[group]
        else:
            g = f
        # Make sure this is our format
        sxs_format = g.attrs["sxs_format"]
        if sxs_format not in sxs_formats:
            raise ValueError(
                f'The `sxs_format` found in H5 file is "{sxs_format}"; it should be one of\n'
                f"    {sxs_formats}."
            )

        # Ensure that the 'validation' keys from the JSON file are the same as in this file
        n_times = g.attrs["n_times"]
        json_n_times = json_data.get("validation", {}).get("n_times", 0)
        if json_n_times != n_times:
            invalid(
                f"\nNumber of time steps in H5 file ({n_times}) "
                f"does not match expected value from JSON ({json_n_times})."
            )

        # Read the raw data
        sizeof_float = 8
        sizeof_complex = 2 * sizeof_float
        ell_min = g.attrs["ell_min"]
        ell_max = g.attrs["ell_max"]
        shuffle_widths = tuple(g.attrs["shuffle_widths"])
        unshuffle = scri.utilities.multishuffle(shuffle_widths, forward=False)
        n_modes = ell_max * (ell_max + 2) - ell_min ** 2 + 1
        i1 = n_times * sizeof_float
        i2 = i1 + n_times * sizeof_complex * n_modes
        uncompressed_data = bz2.decompress(g["data"][...])
        t = np.frombuffer(uncompressed_data[:i1], dtype=np.uint64)
        data = np.frombuffer(uncompressed_data[i1:i2], dtype=np.uint64)
        log_frame = np.frombuffer(uncompressed_data[i2:], dtype=np.uint64)

    # Unshuffle the raw data
    t = unshuffle(t)
    data = unshuffle(data)
    log_frame = unshuffle(log_frame)

    # Reshape and re-interpret the data
    t = t.view(np.float64)
    data = data.reshape((-1, n_times)).T.copy().view(complex)
    log_frame = log_frame.reshape((-1, n_times)).T.copy().view(np.float64)

    # Un-XOR the data
    t = unxor(t)
    data = unxor(data)
    log_frame = unxor(log_frame)

    frame = np.exp(quaternion.as_quat_array(np.insert(log_frame, 0, 0.0, axis=1)))

    w = scri.WaveformModes(
        t=t,
        frame=frame,
        data=data,
        frameType=scri.Corotating,
        dataType=dataType,
        m_is_scaled_out=m_is_scaled_out,
        r_is_scaled_out=r_is_scaled_out,
        ell_min=ell_min,
        ell_max=ell_max,
    )
    w.convert_from_conjugate_pairs()
    w.json_data = json_data

    return w, log_frame
